# cd /Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/phenospy
# /Users/taravser/opt/anaconda3/envs/PhenoScript/bin/python

# ---- new imports
# from pathlib import Path
import os

from phenospy.owl_xmlToOwl import *
from phenospy.phs_addXmlMeta import *
from phenospy.phs_grammar import *
from phenospy.phs_various_fun import *
from phenospy.phs_xml import *
from phenospy.phs_xmlTranslateTerms import *
from phenospy.snips_makeFromYaml import *
from phenospy.snips_readFromJSON import *

# -----------------------------------------
# Converts phs file to owl
# -----------------------------------------

# -----------------------------------------
# ARGUMENTS
# -----------------------------------------
# phs_file    = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/examples/Gryonoides/phs/gryo.phs'
# yaml_file   = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/phenospy/package-data/phs-config.yaml'
# save_dir    = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/examples/'
# save_pref   = 'gryo'

def phsToOWL(phs_file, yaml_file, save_dir, save_pref):
    #
    xml_save = os.path.join(save_dir, save_pref + '.xml')
    owl_save = os.path.join(save_dir, save_pref + '.owl')
    # ----------------------------------------
    # Parse phs file
    # -----------------------------------------
    out = parsePHS(phs_file)
    # -----------------------------------------
    # phs to XML
    # -----------------------------------------
    xml_str, base_iri = phenoscriptToXML(out)
    xml_final = xmlTranslateTermsMeta(xml_str, base_iri, yaml_file)
    # -----------------------------------------
    # Save XML
    # -----------------------------------------
    print(f"{Fore.BLUE}Saving XML...{Style.RESET_ALL}")
    with open(xml_save, 'w') as f:
        f.write(xml_final)
        f.close()
    print(f"{Fore.GREEN}Done! XML is saved to:{Style.RESET_ALL}",  xml_save)
    # -----------------------------------------
    # XML to OWL
    # -----------------------------------------
    # tree = ET.parse(xml_save)
    tree = ET.ElementTree(ET.fromstring(xml_final))
    # set_log_level(0)
    xmlToOwl(tree, owl_save)


# -----------------------------------------------------------
# classes and functions to convert pyparsing output into xml
#
# (C) 2023 Sergei Tarasov
# email sergei.tarasov@helsinki.fi
# -----------------------------------------------------------
import xml.dom.minidom as xmldom
from datetime import datetime
from sys import exit

from phenospy.phs_parser_fun import *
from phenospy.phs_various_fun import *
from phenospy.snips_fun import *

# ------ This is the wrapper to make xml output nice

# former: def phenoscriptParse(phs_string, base_iri):
def phenoscriptToXML(phs_string):
    print(f"{Fore.BLUE}Converting Phenoscript to XML ...{Style.RESET_ALL}")
    # IRI for node: https://urn:uuid:4831b6e2-c326-11ed-a65d-acde48001122/_b3cbc5_12-5
    # IRI for nested node: https://urn:uuid:1312852a-c328-11ed-949a-acde48001122/_ecd5db_14_N3-3"
    # IRI for list node: https://urn:uuid:826c3694-c32a-11ed-96cb-acde48001122/_f34012_15-L3-1
    # For not personalized nodes: https://<base_iri>/<nid>-<self.coord_node>
    # For personalized nodes:     https://<base_iri>/node_dict['id_props']
    #   where node_dict['id_props'] is the tag, N:tag

    # -----------------------------------------
    # Make base iri. This function returns it
    # -----------------------------------------
    # https://urn:uuid:2348c8b6-c31c-11ed-876a-acde48001122
    # base_iri = 'https://' + 'urn:uuid:' + str(uuid.uuid1())
     # https://2348c8b6-c31c-11ed-876a-acde48001122
    base_iri = 'https://' +  str(uuid.uuid1())

    # identifier for to keep track on ophu statement and node coordinates
    global nid # _b3cbc5_N
    base_iri_labels = '_' + uuid_n(6) + '_'  # _b3cbc5_
    nid = nodeIdGenerator(prefix=base_iri_labels, starting_id=1)

    # Global var for keeping track on negative edges
    global varGlobalfromNegativeEdge
    varGlobalfromNegativeEdge = GlobalfromNegativeEdge()

    # IRIs for individuals: "https://urn:uuid:4831b6e2-c326-11ed-a65d-acde48001122/_b3cbc5_12-5"
    global phs_indv_iri
    phs_indv_iri = base_iri + '/'

    nn = '<Phenoscript xmlns:phs="https://github.com/sergeitarasov/PhenoScript">\n' + \
         makeNodes(phs_string[0], pos=1) + '</Phenoscript> '

    xml = xmldom.parseString(nn)
    xml_pretty_str = xml.toprettyxml(indent="\t", newl="\n", encoding=None)
    xx = "\n".join([ll.rstrip() for ll in xml_pretty_str.splitlines() if ll.strip()])
    print(f"{Fore.GREEN}Done!{Style.RESET_ALL}")
    return xx, base_iri


# varGlobalfromNegativeEdge=GlobalfromNegativeEdge()
# varGlobalfromNegativeEdge.updatefromNegativeEdge(True)
class GlobalfromNegativeEdge:
    def __init__(self):
        self.value = 'NA'

    def updatefromNegativeEdge(self, value):
        self.value = value

    def __repr__(self):
        return str(self.value)


# ---------

class phsNode:
    def __init__(self, node_dict, triple_pos='NA', ophu_id='', coord_node='NA', fromNegativeEdge=False):
        self.node_name = node_dict['node_name']
        self.triple_pos = triple_pos  # can be 1,2,3
        # self.node_numeric = False
        self.coord_node = coord_node
        # fromNegativeEdge implies wther the edge that connects to this node has negative prop. assertion
        # this neded for presence/absence reasoning
        self.fromNegativeEdge = fromNegativeEdge

        # node props N[]
        if 'node_props' in node_dict:
            self.props_list = node_dict['node_props']
            self.props_list.insert(0, ['node_name', self.node_name])
        else:
            self.props_list = [['node_name', self.node_name]]

        # id_props is N:xxx, this personalization tag
        if 'id_props' in node_dict:
            self.node_id = phs_indv_iri + node_dict['id_props']
        else:
            # self.node_id = phsIndvIRI + node_id
            self.node_id = phs_indv_iri + str(ophu_id) + '-' + str(self.coord_node)

    def updateNodeID(self, value):
        self.node_id = value

    def nodePropsToXML(self):
        props_collection = []
        # i=1
        for i in range(0, len(self.props_list)):
            coll = '<phs:node_property phs:value="%s">%s</phs:node_property>\n' % (
                str(self.props_list[i][1]), self.props_list[i][0])
            props_collection.append(coll)
        props_collection = ''.join(props_collection)
        # add node name prop
        # props_collection='<phs:node_property phs:value="'+ self.node_name + '">name</phs:node_property>\n' + props_collection
        return props_collection

    def updateTriplePos(self, value):
        self.triple_pos = value

    def strXML(self):
        props_xml = self.nodePropsToXML()
        return '<phs:node phs:node_id="%s" phs:triple_pos="%s" phs:node_name="%s" phs:fromNegativeEdge="%s">\n%s</phs:node>\n' % (
            self.node_id, self.triple_pos, self.node_name, self.fromNegativeEdge, props_xml)

    def __repr__(self):
        return 'node_name:%s  node_id:%s' % (self.node_name, self.node_id)


# ------------
class phsNumericNode:
    def __init__(self, node_dict, triple_pos='NA'):
        self.node_name = node_dict['node_name']
        self.triple_pos = triple_pos  # can be 1,2,3

        # node props N[]
        if 'node_props' in node_dict:
            self.props_list = node_dict['node_props']
            self.props_list.insert(0, ['node_name', self.node_name])
        else:
            self.props_list = [['node_name', self.node_name]]

        # numeric node types e.g., 3.2 or 3
        if 'num_real' in node_dict:
            self.props_list.append(['numeric_type', 'real'])
            self.numeric_type = 'real'

        if 'num_int' in node_dict:
            self.props_list.append(['numeric_type', 'int'])
            self.numeric_type = 'int'

    def nodePropsToXML(self):
        props_collection = []
        # i=1
        for i in range(0, len(self.props_list)):
            coll = '<phs:node_property phs:value="%s">%s</phs:node_property>\n' % (
                str(self.props_list[i][1]), self.props_list[i][0])
            props_collection.append(coll)
        props_collection = ''.join(props_collection)
        return props_collection

    def updateTriplePos(self, value):
        self.triple_pos = value

    def strXML(self):
        props_xml = self.nodePropsToXML()
        return '<phs:numeric_node phs:triple_pos="%s" phs:node_name="%s" phs:numeric_type="%s">\n%s</phs:numeric_node>\n' % (
            self.triple_pos, self.node_name, self.numeric_type, props_xml)
        # return  '<phs:node phs:node_id="%s" phs:triple_pos="%s" phs:node_name="%s" phs:node_numeric="%s">\n%s</phs:node>\n' % (self.node_id, self.triple_pos, self.node_name, self.node_numeric, props_xml)

    def __repr__(self):
        return 'node_name:%s  node_id:%s' % (self.node_name, self.node_id)

# --------

# Quoted Node
class phsQuotedNode:
    def __init__(self, node_dict, triple_pos='NA'):
        self.node_name = node_dict['node_name']
        self.node_name = eval(self.node_name)  # removing double quotes
        self.triple_pos = triple_pos  # should be only 3
        if (self.triple_pos != 3):
            print(f"{Fore.RED}Error: a quoted node should be at triple position 3.{Style.RESET_ALL}")
            print(f"{Fore.RED}\tsome quoted node is at:{Style.RESET_ALL}", self.triple_pos)
            exit()

    def updateTriplePos(self, value):
        self.triple_pos = value

    def strXML(self):
        return '<phs:quoted_node phs:triple_pos="%s" phs:node_name="%s" />\n' % \
            (self.triple_pos, self.node_name)

    def __repr__(self):
        return 'node_name:%s  node_id:%s' % (self.node_name, self.node_id)


# ------
class phsEdge:
    def __init__(self, node_dict, triple_pos='NA'):
        self.node_name = node_dict['edge_name']
        self.triple_pos = triple_pos  # can be 1,2,3
        if 'negative_prop' in node_dict:
            self.negative_prop = 'True'
        else:
            self.negative_prop = 'False'

    def updateNegProp(self, value):
        self.negative_prop = value

    def updateTriplePos(self, value):
        self.triple_pos = value

    def strXML(self):
        # return '<phs:edge phs:triple_pos="%s" phs:negative_prop="%s">%s</phs:edge>\n' % (self.triple_pos, self.negative_prop, self.node_name)
        return '<phs:edge phs:triple_pos="%s" phs:negative_prop="%s" phs:edge_name="%s" />\n' % (
            self.triple_pos, self.negative_prop, self.node_name)

    def __repr__(self):
        return 'edge_name:%s' % (self.node_name)


# ----------------
class phsListNode:
    def __init__(self, node_dict, triple_pos='NA'):
        self.node_name = node_dict['node_name']
        self.triple_pos = triple_pos  # can be 1,2,3

    def updateTriplePos(self, value):
        self.triple_pos = value

    def strXML(self):
        return '<phs:list_node phs:triple_pos="%s"> %s </phs:list_node>\n' % (self.triple_pos, self.node_name)

    def __repr__(self):
        return 'node_name:%s' % (self.node_name)


# --------------
class phsNestedNode:
    def __init__(self, node_dict, triple_pos='NA'):
        self.node_name = 'node_nested'
        self.triple_pos = triple_pos  # can be 1,2,3

    def updateTriplePos(self, value):
        self.triple_pos = value

    def strXML(self):
        return '<phs:node_nested phs:triple_pos="%s"> %s </phs:node_nested>\n' % (self.triple_pos, self.node_name)

    def __repr__(self):
        return 'node_name:%s  node_id:%s' % (self.node_name, self.node_id)


# ------------- This functon parses pyparser ourput into xml

# makeNodes(phs_string[0], pos=1)
# print(phs_string[0][0][1][0])
# node=phs_string[0][0][1][0]
# node=node[0]
# node=node[5]
# makeNodes(node, pos=1)
# makeNodes(phs_string[0][0][1][0], pos=1)
# type(node_dict)
# node=phs_string[0][0][0][0][0]
# node=out[1][0][1][0]
# edge_xml=phsEdge(out[0][0][0][0][0][3])
# edge_xml.negative_prop

def makeNodes(node, pos, ophu_id='NA', coord='NA'):
    positional_dict = {'1': 1, '3': 2, '5': 3}

    if (node._ParseResults__name == 'edge' and isinstance(node, ParseResults)):
        # This is an edge
        node_dict = node.asDict()
        # node_xml=phsEdge(node_dict,  triple_pos=2)
        edge_xml = phsEdge(node_dict, triple_pos=2)
        # print(edge_xml.negative_prop)
        varGlobalfromNegativeEdge.updatefromNegativeEdge(edge_xml.negative_prop)
        return edge_xml.strXML()

    elif (node._ParseResults__name == 'node' and isinstance(node, ParseResults)):
        # This is a node that can be normal or numeric
        node_dict = node.asDict()

        # this is a numeric node for data properties
        if ('num_real' in node_dict) | ('num_int' in node_dict):
            node_xml = phsNumericNode(node_dict, triple_pos=pos)

        # this is quoted node
        elif ('node_quoted' in node_dict):
            node_xml = phsQuotedNode(node_dict, triple_pos=pos)

        # this is normal node
        else:
            # node_xml=phsNode(node_dict,  triple_pos=pos, node_id=nid.makeId())
            node_xml = phsNode(node_dict, triple_pos=pos, ophu_id=ophu_id, coord_node=coord,
                               fromNegativeEdge=varGlobalfromNegativeEdge)
        return node_xml.strXML()

    elif (node._ParseResults__name == 'node_nested' and isinstance(node, ParseResults)):
        # This is a nested node
        out_colection = []
        # i=0
        for i in range(0, len(node)):
            # coordimnates for nodes (e.g., _ophu_6_N5_5) within nested_node consist of: id_ophu_statement / N / coords of the nested node within ophu_statement (coord) / 
            # / coords of node within nested node (assigned by phsNode).
            nested_coord = ophu_id + '_N' + str(coord)
            # print('nested_coord ' + str(nested_coord))
            out_colection.append(makeNodes(node[i], pos='NO', ophu_id=nested_coord))
        out_colection = ''.join(out_colection)
        # '<phs:node_nested phs:triple_pos="%s"> %s </phs:node_nested>\n' % (self.triple_pos, self.node_name)
        out_colection = '<phs:nested_node phs:triple_pos="%s">\n%s</phs:nested_node>\n' % (str(pos), out_colection)
        return out_colection

    elif (node._ParseResults__name == 'list_node' and isinstance(node, ParseResults)):
        # This is a list node
        out_colection = []
        # i=0
        for i in range(0, len(node)):
            # coordimnates for nodes (e.g., _ophu_4_L9_2) within lis_node consist of: id_ophu_statement / L / coords of the list node within ophu_statement (coord) / 
            # coords of the node within the listed node
            # list_coord=str(coord)+'L'+str(i+1)
            list_coord = 'L' + str(coord) + '-' + str(i + 1)
            out_colection.append(makeNodes(node[i], pos='NO', ophu_id=ophu_id, coord=list_coord))
        out_colection = ''.join(out_colection)
        out_colection = '<phs:list_node phs:triple_pos="%s">\n%s</phs:list_node>\n' % (str(pos), out_colection)
        return out_colection

    elif (isinstance(node, TripleNode) and isinstance(node, ParseResults)):
        # This is a triple (N E N)
        out_colection = []
        varGlobalfromNegativeEdge.updatefromNegativeEdge(False)
        # i=5
        for i in [1, 3, 5]:
            # out_colection.append(makeNodes(node[i], pos=positional_dict[str(i)] ))
            # node[i-1]['coord']
            out_colection.append(
                makeNodes(node[i], pos=positional_dict[str(i)], ophu_id=ophu_id, coord=node[i - 1]['coord']))
        out_colection = ''.join(out_colection)
        out_colection = '<phs:triple_node>\n%s</phs:triple_node>\n' % (out_colection)
        return out_colection

    elif (not bool(node._ParseResults__name) and not isinstance(node, TripleNode) and isinstance(node, ParseResults)):
        # This is the collection of triples
        out_colection = []
        # i=1
        for i in range(0, len(node)):
            # print(i)
            out_colection.append(makeNodes(node[i], pos='NO', ophu_id=ophu_id))
        out_colection = ''.join(out_colection)
        return out_colection

    elif (not isinstance(node, TripleNode) and isinstance(node, ParseResults)):
        # This is any other node
        out_colection = []
        # out_colection= node[0]
        xml_element = str(node._ParseResults__name)

        # ----
        ophu_stat_id = ''
        # each ophu_statement gets an id to use for naming individuals
        if (xml_element == "ophu_statement"):
            ophu_stat_id = nid.makeId()
            # print(ophu_stat_id)
        # ----
        # i=0
        # node=node[0]
        for i in range(0, len(node)):
            out_colection.append(makeNodes(node[i], pos='NO', ophu_id=ophu_stat_id))

        out_colection = ''.join(out_colection)
        out_colection = '<phs:' + xml_element + '>\n' + out_colection + '</phs:' + xml_element + '>\n'
        # out_colection='<phs:‰s>\n%s</phs:‰s>\n' % (xml_element, out_colection, xml_element)
        return out_colection

    else:
        print("makeNodes(): INVALID INPUT")



from owlready2 import *
from colorama import Fore, Style
from colorama import init as colorama_init
colorama_init()

# -----------------------------------------
# Helpers
# -----------------------------------------
phenoscript_annotations = IRIS['https://github.com/sergeitarasov/PhenoScript/PHS_0000001']
phs_implies_absence_of  = IRIS['https://github.com/sergeitarasov/PhenoScript/PHS_0000003']
phs_NL                  = IRIS['https://github.com/sergeitarasov/PhenoScript/PHS_0000004']
phs_original_assertion  = IRIS['https://github.com/sergeitarasov/PhenoScript/PHS_0000005']
phs_original_class      = IRIS['https://github.com/sergeitarasov/PhenoScript/PHS_0000002']

phenoscript_annotations.python_name = 'phenoscript_annotations'
phs_implies_absence_of.python_name  = 'phs_implies_absence_of'
phs_NL.python_name                  = 'phs_NL'
phs_original_assertion.python_name  = 'phs_original_assertion'
phs_original_class.python_name      = 'phs_original_class'

# global phenoscript_annotations
# global phs_implies_absence_of 
# global phs_NL                 
# global phs_original_assertion 
# global phs_original_class     

# -----------------------------------------
# Tmp 1: sparql Absolute and Relative Mesuarements
# -----------------------------------------

# obo.PATO_0000122 # length
# obo.RO_0000053 # has_characterristi
# obo.IAO_0000417 # is qual measured as
# obo.IAO_0000004 # has measurement value
# obo.IAO_0000039 # has measurement unit label
# obo.RO_0000052 # characteristic of
# obo.RO_0000053 # has characteristic, bearer of
# obo.BFO_0000051 has part
def sparql_tmpMeasurements(default_world):
    query=list(default_world.sparql("""
            SELECT ?E0 ?Q ?Val ?x ?Unit ?Unit_loc
            WHERE {?E0 obo:RO_0000053 ?Q .
                ?Q obo:IAO_0000417 ?x .
                ?x obo:IAO_0000004 ?Val .
                ?x obo:IAO_0000039 ?Unit .
                optional {
                    ?Unit obo:RO_0000052 ?Unit_loc .
                    }
                }
    """))
    return query


# Structure: dict_templ={"E0":None, "Q":None, "Val":None, "x":None, "Unit":None, "Unit_loc":None}
# [{'E0': org_Grebennikovius_basilewskyi, 'Q': length:_1e96f5_5-3, 'Val': 2.3,
#  'x': measurement datum:md-8ff5d4, 'Unit': millimeter:_1e96f5_5-7, 'Unit_loc': None}]
def tmpMeasureToDic(query):
    sparql=[]
    for item in query:
        # print(item)
        dict_templ = {"E0": None, "Q": None, "Val": None, "x": None, "Unit": None, "Unit_loc": None}
        dict_templ['E0']= item[0]
        dict_templ['Q'] = item[1]
        dict_templ['Val'] = item[2]
        dict_templ['x'] = item[3]
        dict_templ['Unit'] = item[4]
        if len(item)==6:
            dict_templ['Unit_loc'] = item[5]
        sparql.append(dict_templ)
    return sparql

# node to mardown used for tmp Measure
# def getLabelPhsAnnotation(ind):
def nodeToNL(ind):
    if ind is not None:
        return "[%s](%s)" % (ind.phs_original_class[0].label.first(), ind.phs_original_class[0].iri)

# Unit Locator to nl. used for tmp Measure
# measurement datum:md-8ff5d4, 'Unit': millimeter:_1e96f5_5-7, 'Unit_loc': None
# def render_Unit_loc(ind):
def UnitLocToNL(ind):
    if ind is not None:
        return " of [%s](%s);" % (ind.phs_original_class[0].label.first(), ind.phs_original_class[0].iri)
    else:
        return ";"

# dic_mesuare obtained from sparql query to NL in OWL
def dic_mesuare_ToNLinOWL(dic_mesuare):
    for item in dic_mesuare:
        txt = ", %s = %s, unit: %s%s" % \
        ( nodeToNL(item['Q']),
        item['Val'],
        nodeToNL(item['Unit']),
        UnitLocToNL(item['Unit_loc']) )
        # print(txt)
        item['E0'].phs_NL.append(txt)
        # print("NL:", item['E0'].phs_NL)
        # return(dic_mesuare)
    #
    # DELETE
    # dic_mesuare
    for item in dic_mesuare:
        # destroy_ListEntities([item['Q'], item['Unit']])
        destroy_ListEntities([item['Q']])
        if item['x'] is not None:
            # print(item['x'])
            destroy_entity(item['x'])

# destroy list entitities from sparql query
def destroy_ListEntities(list):
    for entity in list:
        destroy_entity(entity)


# -----------------------------------------
# SPARQL:  Relative Comparison
#   Pattern: Org1 > E1 >> Q1 |comp| Q2 << E2 < Org2
# -----------------------------------------

# Algorithm
# 1. Since sparql here does not quey non-existent props
#   run queries for |>| and |<| separately.
#   First check for the prop existence
# 2. Construct the relative comparison phs_NL only is spp are different.
#   i.e., Org1 != Org2
#   If spp are the same do nothing
# 3. delete inherence_in prop if Q2 inherence in E2: delete inherence_in
#

# comparative_prop = phs_ns.PHS_0000017.iri
def sparql_RelatComp(default_world, comparative_prop):
    # phs_ns.PHS_0000017.iri # has_trait
    # obo.RO_0015007 # increased in magnitude relative to
    # obo.RO_0015008 # deacreased in ...
    # obo.RO_0000053 # bearer of
    has_trait_iri = 'https://github.com/sergeitarasov/PhenoScript/PHS_0000017'
    # !!! ?prop is used on purpose to have acess to the object indownstream scripts
    query = list(default_world.sparql("""
            SELECT DISTINCT ?Q1 ?Org1 ?Q2 ?E2 ?Org2 ?prop
            WHERE {
                ?Q1 a owl:NamedIndividual .
                ?Q2 a owl:NamedIndividual .
                ?Q1  <%s> ?Q2 .
                ?Q1  ?prop ?Q2 .
                ?E2 obo:RO_0000053 ?Q2 .
                ?Org1 <%s> ?Q1 .
                ?Org2 <%s> ?Q2 .
            }
    """ % (comparative_prop, has_trait_iri, has_trait_iri)))
    #print(query)
    return query

# -----------------------------------------
#  q1 |CompProp| q2. This functions removes |CompProp|
# between ind: q1 and q2. Creates a fake ind that is connected to q1
# -----------------------------------------
# CompProp = obo.RO_0015007
# q1 = IRIS['https://urn:uuid:a0f607de-cee2-11ed-a2ca-acde48001122/_1e96f5_13-5']
# q2 = IRIS['https://urn:uuid:a0f607de-cee2-11ed-a2ca-acde48001122/id-185304']
# txt = 'some txt'
def make_fakeInd(q1, q2, CompProp, txt):
    # remove annotation
    phs_original_assertion[q1, CompProp, q2] = []
    # remove CompProp link q1.CompProp
    CompProp[q1] = []
    # make fake individual
    CL = q1.is_a.first()
    fake_ind = CL('https://temp/tmp/')
    fake_ind.iri = 'https://temp/tmp/' + q1.label.first() + '/fake_ind'
    fake_ind.phs_original_class.append(CL)
    fake_ind.phs_NL.append(txt)
    # make a link between q1 and fake_ind
    CompProp[q1] = [fake_ind]
    phs_original_assertion[q1, CompProp, fake_ind] = True


def tmpRelatToDic(query):
    # [{'Q1': length:_1e96f5_13-5, 'Org1': org_Grebennikovius_basilewskyi, 
    #   'Q2': length:id-185304, 'E2': leg:_1e96f5_18-3, 'Org2': org_Grebennikovius}]
    dic_RelatComp=[]
    for item in query:
        # print(item)
        dict_templ = {"Q1": None, "Org1": None, "Q2": None, "E2": None, "Org2": None, "CompProp": None}
        dict_templ['Q1']= item[0]
        dict_templ['Org1'] = item[1]
        dict_templ['Q2'] = item[2]
        dict_templ['E2'] = item[3]
        dict_templ['Org2'] = item[4]
        dict_templ['CompProp'] = item[5]
        dic_RelatComp.append(dict_templ)
    return dic_RelatComp

def dic_Relat_ToNLinOWL(dic_RelatComp):
    # We need to:
    # 1. delete inherence_in prop if Q2 inherence in E2 delete inherence_in
    # 2. construct the relative comparison phs_NL only is spp are different
    # --
    # txt: Q2 of E2 in Org2
    # item = dic_RelatComp[0]
    # renderNL(item['CompProp'], onto)
    for item in dic_RelatComp:
        if (item['Org1'].iri != item['Org2'].iri):
            # txt = " %s of %s in _%s_" % \
            # ( nodeToNL(item['Q2']), nodeToNL(item['E2']), item['Org2'].label.first() )
            txt = " of %s in _%s_;" % \
            (nodeToNL(item['E2']), item['Org2'].label.first())
            # print(txt)
            # substitute Q1 with fake ind that is not connected to Q2
            make_fakeInd(q1=item['Q1'], q2=item['Q2'], 
            CompProp =item['CompProp'], txt=txt)


# -----------------------------------------
# Pattern: chracter: present
# -----------------------------------------
def add_presentTag_ToNLinOWL(default_world):
    global phenoscript_annotations, phs_implies_absence_of, phs_NL 
    global phs_original_assertion, phs_original_class
    # get all pairs: x has part y
    # obo.BFO_0000051 has part
    query_x_hp_y = list(default_world.sparql("""
            SELECT DISTINCT ?y
            WHERE {
                ?x a owl:NamedIndividual .
                ?y a owl:NamedIndividual .
                ?x obo:BFO_0000051 ?y .
                }
    """))
    # Old code
    # presesent_tag=set()
    # for trip in query_x_hp_y:
    #     n1=trip[0] # x
    #     n2=trip[1] # y
    #     has_out_edge=False
    #     # n2.get_properties()
    #     for prop in n2.get_properties():
    #         if not issubclass(prop, phs_linksTraits):
    #             for n3 in prop[n2]:
    #                 asrt = phs_original_assertion[n2, prop, n3]
    #                 print(n2, prop, n3, asrt)
    #                 if len(asrt)>0 or issubclass(prop, phs_NL):
    #                     #if issubclass(prop, owl.ObjectProperty) or issubclass(prop, owl.DatatypeProperty):
    #                     #print(n2, prop, "is", asrt)
    #                     has_out_edge = True
    #     if has_out_edge == False:
    #         print(n2, prop, "is", asrt)
    #         presesent_tag.add(n2)

    # every node that should be marks with 'present' does not have any single node that is phs_original_assertion or phs_NL
    # so if node has those assertions it is not our customer. we look for those that have no.
    # ---
    # 1. from sparql, get all pairs: x has part y
    # 2. for each y get triples: y prop z
    # 3. check if for all props:  y prop z is not phs_original_assertion and phs_NL
    # 4. Select y if all Flase are returned at step 3

    presesent_tag=set()
    for item in query_x_hp_y:
        y = item[0]
        has_out_edges = []
        # get all props from y
        for prop in y.get_properties():
            # get triple: y prop z
            for z in prop[y]:
                is_original_assert = bool(phs_original_assertion[y, prop, z])
                # print(y, prop, z, is_original_assert)
                if (is_original_assert == False) and not issubclass(prop, phs_NL):
                    has_out_edges.append(False)
                else:
                    has_out_edges.append(True)
        # if all props return false for a given y then it's what we need
        if all(not item for item in has_out_edges):
            # print(y)
            presesent_tag.add(y)
    #
    # add 'present' tag to NL in OWL
    for ind in presesent_tag:
        txt = ": present;"
        # print(ind, txt)
        ind.phs_NL.append(txt)



# -----------------------------------------
# Make NL Graph for Md rendering
# -----------------------------------------
def makeNLGraph_basic(onto):
    global phenoscript_annotations, phs_implies_absence_of, phs_NL 
    global phs_original_assertion, phs_original_class
    # -----------------------------------------
    # Namespaces
    # -----------------------------------------
    obo = onto.get_namespace("http://purl.obolibrary.org/obo/")
    phs_ns = onto.get_namespace('https://github.com/sergeitarasov/PhenoScript/')

    # -----------------------------------------
    # Add absence
    # -----------------------------------------
    print(f"{Fore.BLUE}Adding absence traits...{Style.RESET_ALL}")

    # for ind in onto.individuals():
    #     #print(ind)
    #     if (len(ind.phs_implies_absence_of)>0):
    #         # print(ind, "---", ind.phs_implies_absence_of)
    #         abs_class=ind.phs_implies_absence_of[0]
    #         txt=" [%s](%s): absent;" % (abs_class.label.first(), abs_class.iri)
    #         # print(txt)
    #         ind.phs_NL.append(txt)
    
    for ind in onto.individuals():
        #print(ind)
        if (len(ind.phs_implies_absence_of)>0):
            #print(len(ind.phs_implies_absence_of))
            #print(ind, "---", ind.phs_implies_absence_of)
            for index in range(0, len(ind.phs_implies_absence_of)):
                abs_class=ind.phs_implies_absence_of[index]
                #print('\t', abs_class)
                txt=", [%s](%s): absent;" % (abs_class.label.first(), abs_class.iri)
                ind.phs_NL.append(txt)

    # -----------------------------------------
    # Pattern: add chracter prsence
    # -----------------------------------------
    print(f"{Fore.BLUE}Adding presence traits...{Style.RESET_ALL}")
    add_presentTag_ToNLinOWL(default_world)

    # -----------------------------------------
    # SPARQL:  Absolute and Relative Mesuarements
    # -----------------------------------------
    print(f"{Fore.BLUE}Adding Absolute and Relative Mesuarements...{Style.RESET_ALL}")
    
    # check is meserument entities are present, othervise sparql does not work
    is_mesure_present = IRIS['http://purl.obolibrary.org/obo/IAO_0000417']
    if bool(is_mesure_present):
        query_tmpMesuare = sparql_tmpMeasurements(default_world)
        dic_mesuare = tmpMeasureToDic(query_tmpMesuare)
        # dic_mesuare obtained from sparql query to NL in OWL
        dic_mesuare_ToNLinOWL(dic_mesuare)

    # -----------------------------------------
    # SPARQL:  Relative Comparison
    #   Pattern: Org1 > E1 >> Q1 |comp| Q2 << E2 < Org2
    # -----------------------------------------
    print(f"{Fore.BLUE}Adding Relative Comparison traits...{Style.RESET_ALL}")

    # Check existence and if existence render it to NL
    # increased in magnitude relative to
    compProp = obo.RO_0015007
    if bool(compProp):
        query_RelatComp = sparql_RelatComp(default_world, comparative_prop = obo.RO_0015007.iri)
        dic_RelatComp = tmpRelatToDic(query_RelatComp)
        dic_Relat_ToNLinOWL(dic_RelatComp)
    #
    # decreased in magnitude relative to
    compProp = obo.RO_0015008
    if bool(compProp):
        query_RelatComp = sparql_RelatComp(default_world, comparative_prop = obo.RO_0015008.iri)
        dic_RelatComp = tmpRelatToDic(query_RelatComp)
        dic_Relat_ToNLinOWL(dic_RelatComp)
    #
    print(f"{Fore.GREEN}Done!{Style.RESET_ALL}")


# -----------------------------------------
# Traverse graph
# -----------------------------------------

global visited_triples
visited_triples = list()

# def traverseOntoMark_Overleaf
def traverseGraphForNL(onto, ind0, tabs="\t", tab_char="\t", visited_nodes=set()):
    global visited_triples
    global phenoscript_annotations, phs_implies_absence_of, phs_NL 
    global phs_original_assertion, phs_original_class
    #print(visited_triples)
    txt=""
    visited_nodes.add(ind0)
    triples=[]
    # prop = list(ind0.get_properties())[7]
    for prop in ind0.get_properties():
        if not issubclass(prop, owl.AnnotationProperty):
            # print('HERE', prop)
            for value in prop[ind0]:
                asrt = phs_original_assertion[ind0, prop, value]
                triple = [ind0, prop, value]
                if len(asrt) > 0 and (triple not in visited_triples): # should be also True for asrt
                    #triple = [ind0, prop, value]
                    #print(asrt)
                    triples.append(triple)
                    visited_triples.append(triple)
        #elif issubclass(prop, inf.PhenoScript_NL) or issubclass(prop, inf.PhenoScript_implies_absence_of):
        elif issubclass(prop, phs_NL):
            for value in prop[ind0]:
                triple = [ind0, prop, value]
                if (triple not in visited_triples):
                    #triple = [ind0, prop, value]
                    triples.append(triple)
                    visited_triples.append(triple)
    if len(triples)==1:
        prop=triples[0][1]
        value=triples[0][2]
        #txt = txt + (" <%s> %s" % (prop, value))
        txt = txt + ("%s%s" % (renderNL(prop, onto), renderNL(value, onto)))
        if issubclass(prop, owl.ObjectProperty) and (value not in visited_nodes):
            #print(traverseOnto(value))
            #visited_nodes.add(value)
            txt = txt + traverseGraphForNL(onto, value, tabs=tabs, tab_char = tab_char, visited_nodes=visited_nodes)
        return  txt
    if len(triples) > 1:
        # trip = triples[0]
        for trip in triples:
            # trip= triples[0]
            prop = trip[1]
            value = trip[2]
            #txt = txt +"\n" + tabs + ("%s <%s> %s" % (ind0, prop, value))
            txt = txt + "\n" + tabs + ("%s%s%s" % (renderNL(ind0, onto), renderNL(prop, onto), renderNL(value, onto)))
            if issubclass(prop, owl.ObjectProperty) and (value not in visited_nodes):
                # txt = txt + traverseOntoMark_Overleaf(value, tabs=tabs+"\t", visited_nodes=visited_nodes)
                txt = txt + traverseGraphForNL(onto, value, tabs=tabs+tab_char, tab_char = tab_char, visited_nodes=visited_nodes)
        return txt
    if len(triples) == 0:
        # print('TRIPR len=0')
        return ";"



def check_string_pattern_complete(s):
    if (s.startswith(',') or s.startswith(':')):
        return True
    return False

# -----------------------------------------
# Render NL when traversing graph
# -----------------------------------------
# obo.RO_0015007 increased_in_magnitude_relative_to http://purl.obolibrary.org/obo/RO_0015007
# obo.RO_0015008 decreased in magnitude relative to http://purl.obolibrary.org/obo/RO_0015008
def renderNL(x, onto):
    obo = onto.get_namespace("http://purl.obolibrary.org/obo/")
    # if x is individual
    if isinstance(x, owl.Thing):
        return " [%s](%s)" % (x.phs_original_class[0], x.phs_original_class[0].iri)
    elif issubclass(x, owl.ObjectProperty) or issubclass(x, owl.DatatypeProperty):
        #
        dict={obo.BFO_0000051 :',', obo.BFO_0000050 : ' of', obo.RO_0000053 : ':', obo.RO_0000052 : ' of', 
        obo.RO_0015007 : ' [larger than](http://purl.obolibrary.org/obo/RO_0015007)',  
        obo.RO_0015008 : ' [smaller than](http://purl.obolibrary.org/obo/RO_0015008)'}
        #
        if x in dict:
            return dict[x]
        else:
            return " [%s](%s)" % (x.label.first(), x.iri)
    elif issubclass(x, owl.AnnotationProperty):
        return ""
    elif isinstance(x, (int, float)):
        return " %s;" % str(x)
    elif isinstance(x, str):
        #return x
        #return "%s" % x
        if check_string_pattern_complete(x):
           return "%s" % x
        else:
            return  " %s" % x


# -----------------------------------------
# Make NL desciptions in Md format
# -----------------------------------------
# ind0 = onto.search(label = 'org_Grebennikovius_basilewskyi')[0]
# def makeMarkdown(onto, ind0, file_save=None, verbose=False):
#     print(f"{Fore.BLUE}Rendering NL description as markdown...{Style.RESET_ALL}")
#     global visited_nodes, visited_triples
#     visited_nodes=set()
#     visited_triples=list()
#     md = traverseGraphForNL(onto, ind0, tabs='\t', tab_char='\t', visited_nodes=set())
#     # print(md)
#     md_out = md.replace("\t [", " [")
#     md_out = md_out.replace("\t [", "\t- [")
#     md_out = md_out.replace("\n [", "\n- [")
#     # Add species name as comment
#     header = '<!-- ' + ind0.label.first() + ' -->'
#     md_out = header + md_out
#     #print(md_out)
#     #
#     visited_nodes=set()
#     visited_triples=list()
#     #
#     if file_save is not None:
#         print(f"{Fore.BLUE}Saving markdown to:{Style.RESET_ALL}", file_save)
#         fl = open(file_save, "w+")
#         fl.writelines(md_out)
#         fl.close()
#     print(f"{Fore.GREEN}Done!{Style.RESET_ALL}")
#     if verbose==True:
#         return md_out


# -----------------------------------------------------------
# Core Phenoscript phs Parser
#
# (C) 2023 Sergei Tarasov
# email sergei.tarasov@helsinki.fi
# -----------------------------------------------------------
from phenospy.phs_various_fun import *
from pyparsing import *
from pyparsing import pyparsing_common as ppc


# Get odd mubers in range
def oddNum(start, end):
    xx = []
    for num in range(start, end + 1):
        # checking condition
        if num % 2 != 0:  # if odd: num % 2 != 0:
            # print(num, end = " ")
            xx.append(num)
    return xx


oddNum.__doc__ = """
Create odd numbers in a range
     - odd nums
"""


# Get even numbers in range
def evenNum(start, end):
    xx = []
    for num in range(start, end + 1):
        # checking condition
        if num % 2 == 0:  # if odd: num % 2 != 0:
            # print(num, end = " ")
            xx.append(num)
    return xx


# Group ophu elements into triplets and mark them with position
# groupTriplet(tokens=['n1', 'e1', 'n2@h1', 'e2', 'n3', 'e3', 'n4'])
# make a class for triples to recognize in the pyparser output

class TripleNode(ParseResults):
    pass

def groupTriplet(tokens):
    out = []
    # we work with even nums since enumeration in python starts with 0
    even_start = evenNum(0, len(tokens))
    even_end = evenNum(0, len(tokens))
    del even_start[-1]
    del even_end[0]
    # print(even_start)
    # print(even_end)
    # i=0
    # print(tokens)
    for i in range(0, len(even_start)):
        triple = tokens[even_start[i]:(even_end[i] + 1)]

        # ---- check for errors. triple should be N E N. if not report an error
        # print('triple: ', triple[0]._ParseResults__name, triple[1]._ParseResults__name, triple[2]._ParseResults__name)
        # triple_entities=['node', 'edge', 'node']
        triple_entities = [triple[0]._ParseResults__name, triple[1]._ParseResults__name, triple[2]._ParseResults__name]
        trpl0 = (triple_entities[0] not in {'node', 'node_nested', 'list_node'})
        trpl1 = (triple_entities[1] not in {'edge'})
        trpl2 = (triple_entities[2] not in {'node', 'node_nested', 'list_node'})

        # if (triple[0]._ParseResults__name!='node' or triple[1]._ParseResults__name!='edge' or triple[2]._ParseResults__name!='node'  ):
        if (trpl0 or trpl1 or trpl2):
            # print('ERROR')
            return ('ERROR')
        # ---

        # triple=['pos=1',triple[0], 'pos=2', triple[1], 'pos=3', triple[2] ]
        triple = [{'pos': 1, 'coord': even_start[i] + 1}, triple[0], {'pos': 2, 'coord': even_start[i] + 2}, triple[1],
                  {'pos': 3, 'coord': even_start[i] + 3}, triple[2]]
        triple = TripleNode(triple)
        # type(triple)
        out.append(triple)
        # out.extend(triple)
    return out


def transform4(string, location, tokens):
    out = groupTriplet(tokens)
    if (out == 'ERROR'):
        print('ERROR: the triplet syntax is not Node Edge Node, see line:', txt[1:location].count('\n') + 1,
              ' (location:', location, ')')
        exit('Syntax ERROR found')
    return out


def make_keyword(kwd_str, kwd_value):
    return Keyword(kwd_str).setParseAction(replaceWith(kwd_value))


# count read OTU objects
# var_countOTU=nodeIdGenerator(prefix='', starting_id=1)

global var_countOTU
var_countOTU = nodeIdGenerator(prefix='', starting_id=1)

# string='sdcc \n dsdc \n dsd \n'
# location=7
def countOTUobjects(string, location, tokens):
    # print('Reading:', string[location:(location+30)], '\n')
    # substr = string[location:len(string)]
    ednOfLine = string[location:len(string)].find('\n')
    ednOfLine = ednOfLine + location
    # print('- Reading', var_countOTU.makeId(), ':', string[location:ednOfLine], flush=True)
    print(f"{Fore.GREEN}- Reading OTU",  var_countOTU.makeId(), ':', f"{Style.RESET_ALL}", string[location:ednOfLine])
    return tokens





from colorama import Fore, Style
from colorama import init as colorama_init
colorama_init()
from phenospy.owl_owlready_config import *
import re

# -----------------------------------------
# Make NL Graph for Md rendering from OWL: high-level wrapper
# -----------------------------------------
def owlToNLgraph(owl_file):
    # -----------------------------------------
    # Arguments
    # -----------------------------------------
    # set_log_level(0)
    print(f"{Fore.BLUE}Reading OWL:{Style.RESET_ALL}", owl_file)
    onto = get_ontology(owl_file).load(reload_if_newer=True, reload=True)
    # -----------------------------------------
    # Namespaces
    # -----------------------------------------
    obo = onto.get_namespace("http://purl.obolibrary.org/obo/")
    phs_ns = onto.get_namespace('https://github.com/sergeitarasov/PhenoScript/')
    import phenospy.nl_fun
    # -----------------------------------------
    # Make NL graph 
    # -----------------------------------------
    phenospy.nl_fun.makeNLGraph_basic(onto)
    return onto


def substitute_unwanted_combinations(string):
    modified_string = string.replace(" ,", ",")
    modified_string = modified_string.replace(" :", ":")
    return modified_string

# finds lines that belong to metadata and placces them first
def find_and_move_lines(huge_string):
    # pattern = r'\)\s\['
    pattern = '^- [^\]]+\]\([^\)]+\) [^\]]+\]\([^\)]+\)'
    #lines = huge_string.split('\n')
    lines=re.split(r'\n(?!\t)', huge_string)
    matched_lines = []

    for line in lines:
        if re.search(pattern, line):
            matched_lines.append(line)

    modified_string = '\n'.join(matched_lines) + '\n' + '---' + '\n'.join(line for line in lines if line not in matched_lines)

    return modified_string

# -----------------------------------------
# Make NL desciptions in Md format
# -----------------------------------------
# ind0 = onto.search(label = 'org_Grebennikovius_basilewskyi')[0]
def NLgraphToMarkdown(onto, ind0, file_save=None, verbose=False):
    
    print(f"{Fore.BLUE}Converting NL graph to Markdown...{Style.RESET_ALL}")
    global visited_nodes, visited_triples
    visited_nodes=set()
    visited_triples=list()
    #
    import phenospy.nl_fun
    md = phenospy.nl_fun.traverseGraphForNL(onto, ind0, tabs='\t', tab_char='\t', visited_nodes=set())
    # print(md)
    md_out = md.replace("\t [", " [")
    md_out = md_out.replace("\t [", "\t- [")
    md_out = md_out.replace("\n [", "\n- [")
    # remove unwanted combbinations of characters
    md_out = substitute_unwanted_combinations(md_out)
    # finds lines that belong to metadata and placces them first
    md_out = find_and_move_lines(md_out)
    # Add species name as comment
    header = '<!-- ' + ind0.label.first() + ' -->\n'
    md_out = header + md_out
    #print(md_out)
    #
    visited_nodes=set()
    phenospy.nl_fun.visited_triples=list()

    
    if file_save is not None:
        print(f"{Fore.BLUE}Saving Markdown to:{Style.RESET_ALL}", file_save)
        fl = open(file_save, "w+")
        fl.writelines(md_out)
        fl.close()
    print(f"{Fore.GREEN}Done!{Style.RESET_ALL}")
    if verbose==True:
        return md_out




from phenospy.snips_fun import *


# -----------------------------------------
# Make snippets for VScode
# -----------------------------------------
# yaml_file = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/phenospy/package-data/phs-config.yaml'
# make_vscodeSnips(yaml_file)
def make_vscodeSnips(yaml_file):
    # -----------------------------------------
    # Read configuration yaml
    # -----------------------------------------

    print(f"{Fore.BLUE}Reading yaml file...{Style.RESET_ALL}")
    with open(yaml_file, 'r') as f_yaml:
        phs_yaml = yaml.safe_load(f_yaml)
    # print(phs_yaml)
    print(f"{Fore.GREEN}Good! File is read!{Style.RESET_ALL}")

    # -----------------------------------------
    # Processing yaml namespace
    # ----------------------------------------

    nmp_store=[]
    nmp = phs_yaml['namespace']['others']
    for key in nmp:
        # [IRI length, prefix, IRI, is.default.namespace]
        x = [len(nmp[key]), key, nmp[key], False ]
        nmp_store.append(x)
        #print(x)

    # add default
    deff = phs_yaml['namespace']['default']
    nmp_store.append([len(deff), 'Default', deff, True ])
    nmp_store.sort(reverse=True)

    print(f"{Fore.GREEN}Good! Yaml namespace is processed!{Style.RESET_ALL}")

    # -----------------------------------------
    # Check if namespace has unique elements
    # -----------------------------------------
    prefs = [i[1] for i in nmp_store]
    iris = [i[2] for i in nmp_store]

    dupl = set([x for x in prefs if prefs.count(x) > 1])
    if len(dupl)>0:
        print(f"{Fore.RED}Warning! Yaml contains  duplicated namespace prefixes. Fix it!{Style.RESET_ALL}")
        exit()
    dupl = set([x for x in iris if iris.count(x) > 1])
    if len(dupl)>0:
        print(f"{Fore.RED}Warning! Yaml contains  duplicated namespace IRIs. Fix it!{Style.RESET_ALL}")
        exit()

    # -----------------------------------------
    # Processing ontologies
    # -----------------------------------------

    # def render_using_label(entity):
    #     return entity.label.first() or entity.name
    #
    # set_render_func(render_using_label)
    # set_log_level(9)

    print(f"{Fore.BLUE}Start reading ontologies...{Style.RESET_ALL}")

    # get ontologies from yaml
    onto_store = phs_yaml['importOntologies']
    # keep all snips here:
    snips = []
    for onto_path in onto_store:
        # print(onto_path)
        print(f"\n\t{Fore.BLUE}Loading ontology:{Style.RESET_ALL}", onto_path)
        try:
            onto = get_ontology(onto_path).load()
        except:
            onto = get_ontology(onto_path).load()
        obo = onto.get_namespace("http://purl.obolibrary.org/obo/")
        print(f"\t\t{Fore.GREEN}The ontology loaded!{Style.RESET_ALL}")
        # -----------------------------------------
        # Making snippets
        # -----------------------------------------
        # --------------
        # Classes (C)
        # --------------
        print(f"\t{Fore.BLUE}Extracting snippets...{Style.RESET_ALL}")
        #
        onto_obj = onto.classes()
        for item in onto_obj:
            # print(item.iri)
            # print(item.IAO_0000115.first()) # defintion
            #print(item, item.iri, getNamespace(item.iri, nmp_store))
            pref, trimmed_iri = getNamespace(item.iri, nmp_store)
            if (pref is not None) and (trimmed_iri is not None):
                try:
                    defin = item.IAO_0000115.first()
                except:
                    defin = ''
                if defin is None:
                    defin=''
                snip = snippetInfo('C', item.label.first(), item.iri, pref, trimmed_iri, defin)
                snips.append(snip)
                #print(snip)
            else:
                print(f"\t\t\t{Fore.RED}Warning! See phs-config.yaml. Namespace is not defined for{Style.RESET_ALL}{Fore.BLUE}", item, item.iri)
        # --------------
        # OP
        # --------------
        onto_obj = onto.object_properties()
        for item in onto_obj:
            #print(item.iri)
            # print(item.IAO_0000115.first()) # defintion
            #print(item, item.iri, getNamespace(item.iri, nmp_store))
            pref, trimmed_iri = getNamespace(item.iri, nmp_store)
            if (pref is not None) and (trimmed_iri is not None):
                defin = item.IAO_0000115.first()
                if defin is None:
                    defin=''
                snip = snippetInfo('OP', item.label.first(), item.iri, pref, trimmed_iri, defin)
                snips.append(snip)
                #print(snip)
            else:
                print(f"\t\t\t{Fore.RED}Warning! See phs-config.yaml. Namespace is not defined for{Style.RESET_ALL}{Fore.BLUE}", item, item.iri)
        # --------------
        # DP
        # --------------
        onto_obj = onto.data_properties()
        for item in onto_obj:
            pref, trimmed_iri = getNamespace(item.iri, nmp_store)
            if (pref is not None) and (trimmed_iri is not None):
                defin = item.IAO_0000115.first()
                if defin is None:
                    defin=''
                snip = snippetInfo('DP', item.label.first(), item.iri, pref, trimmed_iri, defin)
                snips.append(snip)
                #print(snip)
            else:
                print(f"\t\t\t{Fore.RED}Warning! See phs-config.yaml. Namespace is not defined for{Style.RESET_ALL}{Fore.BLUE}", item, item.iri)
        # --------------
        # AP
        # --------------
        onto_obj = onto.annotation_properties()
        for item in onto_obj:
            pref, trimmed_iri = getNamespace(item.iri, nmp_store)
            if (pref is not None) and (trimmed_iri is not None):
                defin = item.IAO_0000115.first()
                if defin is None:
                    defin=''
                snip = snippetInfo('AP', item.label.first(), item.iri, pref, trimmed_iri, defin)
                snips.append(snip)
                #print(snip)
            else:
                print(f"\t\t\t{Fore.RED}Warning! See phs-config.yaml. Namespace is not defined for{Style.RESET_ALL}{Fore.BLUE}", item, item.iri)
        #
        print(f"\t{Fore.GREEN}The snippets extracted!\n{Style.RESET_ALL}")

    #len(snips)


    # -----------------------------------------
    # Select Unique IRIs
    # -----------------------------------------

    # get all iris from snips
    iris_phs=[]
    for snip in snips:
        #print(snip.label_phs)
        iris_phs.append(snip.iri)

    # select indices of unique iris
    iri_index = list(set(i for i, x in enumerate(iris_phs) if x not in iris_phs[:i]))
    # select unique snips based on the iri indeces
    snips=[snips[i] for i in iri_index]

    iris_phs=[]
    for snip in snips:
        #print(snip.label_phs)
        iris_phs.append(snip.iri)

    if len(iris_phs) == len(set(iris_phs)):
        print(f"\t{Fore.GREEN}Good! All snippet IRIs are unique!{Style.RESET_ALL}")
        print(f"\t{Fore.GREEN}Number of unique snippet IRIs is:{Style.RESET_ALL}", len(iris_phs))
    else:
        print(f"\t{Fore.GREEN}Warning! The snippet IRIs contain duplicates! Fix it!\n{Style.RESET_ALL}")
        exit()

    # -----------------------------------------
    # Duplicated snippet labels
    # -----------------------------------------

    # check duplicates
    label_phs=[]
    for snip in snips:
        #print(snip.label_phs)
        label_phs.append(snip.label_4_translation)

    unique_label_phs = set(label_phs)
    all_dupl=[]
    if (len(label_phs) != len(unique_label_phs)):
        print(f"{Fore.RED}Warning! Some labels are duplicated! {Style.RESET_ALL}")
        print(f"{Fore.BLUE}Resolving duplicated labels! For each duplicate adding <label>_IRI ...{Style.RESET_ALL}")
        for unique_lab in unique_label_phs:
            # print(unique_lab)
            indeces = [index for index, element in enumerate(label_phs) if element == unique_lab]
            if (len(indeces) > 1):
                all_dupl.append(indeces)
        # i=499
        for dupl in all_dupl:
            for i in dupl:
                snips[i].label_phs = snips[i].label_phs + '_' + snips[i].trimmed_iri
                snips[i].label_4_translation = snips[i].label_4_translation + '_' + snips[i].trimmed_iri
                snips[i].label_no_ns = snips[i].label_no_ns + '_' + snips[i].trimmed_iri
        print(f"{Fore.GREEN}Good! All labels are unique!{Style.RESET_ALL}")
    else:
        print(f"{Fore.GREEN}Good! All labels are unique!{Style.RESET_ALL}")


    # -----------------------------------------
    # Check if snippets contains prohibited characters
    # -----------------------------------------

    # check if snippets contains prohibited characters
    prohibitedChars = ['#', '/']
    print(f"{Fore.BLUE}Checking if snippet contain prohibited characters:{Style.RESET_ALL}", prohibitedChars)

    for snip in snips:
        #print(snip)
        x = snip.label_phs
        # if ('#' in x) or ('/' in x):
        if any(char in x for char in prohibitedChars):
            print(f"{Fore.RED}Warning! The snippet contains prohibited characters. Fix it! {Style.RESET_ALL}")
            print(snip)

    # -----------------------------------------
    # Read default snips from JSON
    # -----------------------------------------



    # from pkg_resources import resource_filename
    # filepath = resource_filename('phenospy', 'package-data/snippets-default.json')
    print(f"{Fore.BLUE}Adding the inbuilt snippets...{Style.RESET_ALL}")
    f_path = os.path.join(get_phenospyPath(), "package-data", "snippets-default.json")
    f = open (f_path)
    # f = open('/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/phenospy/package-data/snippets-default.json')
    json_default = json.load(f)
    f.close()

    cc = 1
    for i in json_default:
        print(cc, json_default[i]['prefix'])
        cc = cc +1


    snips_default = json.dumps(json_default, indent=0)[1:-1]
    snips_default = snips_default.replace('\n]', ']')
    snips_default = snips_default.replace('[\n', '[')
    #print(snips_default)

    # -----------------------------------------
    # Save JSON snippets
    # -----------------------------------------
    # combine snippets
    snips_new = ",\n".join([snip.printJson() for snip in snips])
    snips_out = snips_default + ',' + snips_new
    snips_out = '{\n' + snips_out + '\n}'
    # save
    f_out = phs_yaml['snippet-dir']
    f_out = os.path.join(f_out, "phs-snippets.json")
    print(f"{Fore.BLUE}Saving snippets to:{Style.RESET_ALL}", f_out)
    # f_out = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/phenospy/package-data/phs-snippets.json'
    with open(f_out, 'w', encoding='utf-8') as f:
        f.write(snips_out)
        f.close()

    print(f"{Fore.GREEN}DONE!{Style.RESET_ALL}")



# -----------------------------------------------------------
# Phenoscript Grammar
#
# (C) 2023 Sergei Tarasov
# email sergei.tarasov@helsinki.fi
# -----------------------------------------------------------
from pathlib import Path

from colorama import Fore, Style
from colorama import init as colorama_init

colorama_init()

from phenospy.phs_parser_fun import *
from pyparsing import *
from pyparsing import pyparsing_common as ppc

LPARAN, RPARAN, LBRACK, RBRACK, LBRACE, RBRACE, COLON, EQAL = map(Suppress, "()[]{}:=")

TRUE = make_keyword("true", True)
FALSE = make_keyword("false", False)
NULL = make_keyword("null", None)

# -----------------------------------------
# Keywords/Aliases
# 1. Aliases (keywords) are translated into _xxx_ alias during parsing
# to avoid broken tokens in xml
# 2. Aliases should be defined in JSON snippets
# see package-data/snippets-default.json
# -----------------------------------------

ed5 = make_keyword("|>|", '_vgv_')
ed6 = make_keyword("|<|", '_vlv_')
ed7 = make_keyword("|=|", '_vev_')

# new keywords
ed8 = make_keyword("->", '_hg_')
ed9 = make_keyword("<-", '_gh_')

ed1 = make_keyword(">", '_g_')
ed3 = make_keyword("<", '_l_')
ed2 = make_keyword('>>', '_gg_')
ed4 = make_keyword('<<', '_ll_')

# -----------------------------------------

# graph_word = Word(alphanums +"_", alphanums+"_")
# jsonNumber = ppc.number()
phsInt = ppc.integer()('num_int')
phsReal = ppc.real()('num_real')

graph_alphanum = Word(alphanums + "_" + "-")

#graph_word = (phsReal | phsInt | graph_alphanum)

# -----------------------------------------
# Node
# -----------------------------------------

# --------- Node Properties: N[]
propertyWord = Word(alphanums + "_" + ":")  # in []
# propertyWord = Suppress('.') + Word(alphanums + "_" + ":" + "." + "-")  # in []
jsonStr = quotedString().setParseAction(removeQuotes)
jsonString = (jsonStr | propertyWord)

# ----
jsonStr2 = quotedString()('node_quoted')
graph_word = (phsReal | phsInt | graph_alphanum | jsonStr2)
# ---

jsonNumber = ppc.number()

jsonValue = (jsonString | jsonNumber | TRUE | FALSE | NULL)
memberDef = Group(jsonString + EQAL + jsonValue)  # ("jsonMember")
props = Group(LBRACK + delimitedList(memberDef, delim=', ', combine=False) + RBRACK)("node_props")

# --------- node id N:
ID_props = Suppress(':') + graph_alphanum('id_props')

# --------- Node
# node=Group(graph_word("node_name") + Optional(ID_props) + Optional(props) + NotAny('.') )("node")
node = Group(NotAny('.') + graph_word("node_name") + Optional(ID_props) + Optional(props))("node")

# -----------------------------------------
# Edge
# -----------------------------------------
neg_prop = Literal('!')("negative_prop")
# edge = Group(Optional(neg_prop) + graph_word("edge_name") + Suppress('.'))("edge")
# ------ NEW
# edge_word = Group(Optional(neg_prop) + graph_word("edge_name") + Suppress('.'))("edge")
edge_word = Group(Optional(neg_prop) + Suppress('.') + graph_word("edge_name"))("edge")
# edge_vals=(ed5 |ed6 |ed7 | ed2 | ed1 | ed4 | ed3)("edge_name")
edge_vals = (ed8 | ed9 | ed5 | ed6 | ed7 | ed2 | ed1 | ed4 | ed3)("edge_name")
edge_sign = Group(Optional(neg_prop) + edge_vals)("edge")
edge = (edge_word | edge_sign)
# -----

# -----------------------------------------
# Nested node (N E N E N)
# -----------------------------------------
nested_node1 = LPARAN + OneOrMore(node | edge) + RPARAN
nested_node1.setParseAction(transform4)
nested_node = Group(nested_node1)("node_nested")

# -----------------------------------------
# List_node (N, N, N)
# -----------------------------------------
list_node = Group(LPARAN + delimitedList(node, delim=', ', combine=False) + RPARAN)(
    "list_node")  # if combine=True <node>ventral_region, s1, s2</node>

grammar = OneOrMore(edge | node | nested_node | list_node)
grammar.setParseAction(transform4)

# -----------------------------------------
# OTU statements
# -----------------------------------------
ophu_stat = Group(grammar + Suppress(";"))("ophu_statement")
# ophu_list = Group(Suppress(Literal('OPHU_LIST')) + EQAL + LBRACE + OneOrMore(ophu_stat) + RBRACE)("ophu_list")
# otu_list = Group(Suppress(Literal('OTU_DATA')) + EQAL + LBRACE + OneOrMore(ophu_stat) + RBRACE)("otu_properties")
# otu_object = Group(Suppress(Literal('OTU')) + EQAL + LBRACE + otu_list + ophu_list + RBRACE)("otu_object")

ophu_list = Group(Suppress(Literal('TRAITS')) + EQAL + LBRACE + OneOrMore(ophu_stat) + RBRACE)("ophu_list")
otu_list = Group(Suppress(Literal('DATA')) + EQAL + LBRACE + OneOrMore(ophu_stat) + RBRACE)("otu_properties")
otu_object = Group(Suppress(Literal('OTU')) + EQAL + LBRACE + otu_list + ophu_list + RBRACE)("otu_object")

otu_object.setParseAction(countOTUobjects)

grammar1 = Group(OneOrMore(otu_object))("otu_objects_list")

# comments are denoted by #
pyComment = pythonStyleComment
grammar1.ignore(pyComment)

# Parse phs string
def parsePHS(file_phs):
    print(f"{Fore.BLUE}Reading Phenoscript file:{Style.RESET_ALL}", file_phs)
    txt = Path(file_phs).read_text()
    print(f"{Fore.GREEN}Good! File is read!{Style.RESET_ALL}")
    # check if brackets are balanced
    is_balanced(txt)
    # Parse
    print(f"{Fore.BLUE}Parsing Phenoscript file ...{Style.RESET_ALL}")
    print(f"{Fore.BLUE}The following OTUs found:{Style.RESET_ALL}")
    out = grammar1.parseString(txt)
    var_countOTU.reset(prefix='', starting_id=1)
    return out




import codecs
from phenospy.snips_fun import *

# -----------------------------------------
# Read VScode snips in JSON and make dictionaries
# -----------------------------------------

# yaml_file = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/phenospy/package-data/phs-config.yaml'
# json_iri, json_type, json_lab = make_dicsFromSnips(yaml_file)
def make_dicsFromSnips(phs_yaml):
    # -----------------------------------------
    # Read configuration yaml
    # -----------------------------------------

    # print(f"{Fore.BLUE}Reading yaml file...{Style.RESET_ALL}")
    # with open(yaml_file, 'r') as f_yaml:
    #     phs_yaml = yaml.safe_load(f_yaml)
    # # print(phs_yaml)
    # print(f"{Fore.GREEN}Good! File is read!{Style.RESET_ALL}")

    # -----------------------------------------
    # Read JSON snippets
    # -----------------------------------------

    f_path = phs_yaml['snippet-dir']
    f_path = os.path.join(f_path, "phs-snippets.json")
    print(f"{Fore.BLUE}Reading snippets from JSON:{Style.RESET_ALL}", f_path)

    with codecs.open(f_path, 'r', encoding='utf-8') as f:
        json_file = json.load(f)
    # f = open(f_path)
    # json_file = json.load(f)
    # f.close()
    print(f"{Fore.GREEN}Good! File is read!{Style.RESET_ALL}")

    # for i in json_file:
    #     print(json_file[i]['prefix'])
    # ss= {}
    # ss.update({'ro-bearer_of': 'http://purl.obolibrary.org/obo/RO_0000053'})
    # ss.update({'>>': 'http://purl.obolibrary.org/obo/RO_0000053'})

    # -----------------------------------------
    # Double check if labels are unique
    # -----------------------------------------
    print(f"{Fore.BLUE}Checking if labels (label_4_translation) in snippets are unique...{Style.RESET_ALL}")
    # check duplicates
    label_phs = []
    for i in json_file:
        if (json_file[i]['type'] != 'Tmp'):
            label_phs.append(json_file[i]['label_4_translation'][0])
    #
    # label_phs.append('pato-swollen')
    unique_label_phs = set(label_phs)
    all_dupl = []
    if (len(label_phs) != len(unique_label_phs)):
        print(f"{Fore.RED}Warning! Some labels are duplicated! Fix it!{Style.RESET_ALL}")
        for unique_lab in unique_label_phs:
            # print(unique_lab)
            indeces = [index for index, element in enumerate(label_phs) if element == unique_lab]
            if (len(indeces) > 1):
                all_dupl.append(indeces)
        # i=499
        for dupl in all_dupl:
            for i in dupl:
                print(label_phs[i])
        exit()
    else:
        print(f"{Fore.GREEN}Good! All snippet labels are unique!{Style.RESET_ALL}")

    # -----------------------------------------
    # Make snippet dictionaries
    # -----------------------------------------

    json_iri = {}       # 'hao-annellus': 'http://purl.obolibrary.org/obo/HAO_0000095'
    json_type = {}      # 'http://purl.obolibrary.org/obo/HAO_0000097': 'C'
    json_lab = {}       # 'http://purl.obolibrary.org/obo/HAO_0000097': 'anteclypeus'
    for i in json_file:
        #print(json_file[i]['label_4_translation'])
        if (json_file[i]['type'] != 'Tmp'):
            lab = json_file[i]['label_4_translation']
            iri = json_file[i]['iri']
            lab_org = json_file[i]['label_original']
            type = json_file[i]['type']
            for la in lab:
                json_iri.update({la: iri})
            json_type.update({iri: type})
            json_lab.update({iri: lab_org})

    print(f"{Fore.GREEN}Good! Dictionaries from JSON snippets are constructed!{Style.RESET_ALL}")
    # json_iri['<']
    # json_iri['bfo-part_of']
    # json_type['http://purl.obolibrary.org/obo/BFO_0000050']
    # json_lab['http://purl.obolibrary.org/obo/BFO_0000050']
    return json_iri, json_type, json_lab


# dic_iri, dic_type, dic_lab = make_dicsFromSnips(yaml_file)
class snipsLabelDictionary:
    def __init__(self, dic_iri, dic_type, dic_lab):
        self.untranslated = []
        self.countNA = 0
        self.countLBS = 0
        self.dic_iri = dic_iri
        self.dic_type = dic_type
        self.dic_lab = dic_lab
        # dt = pd.read_csv(file)
        # key = pd.DataFrame(dt, columns=['IRI', 'label.final'])
        # ll = key.to_dict('records')
        # self.dd = {}
        # for i in ll:
        #     # print(i)
        #     # dd[i['IRI']]=i['label.nospace']
        #     self.dd[i['label.final']] = i['IRI']
    def resetCount(self):
        self.countNA = 0
        self.countLBS = 0
    def translate_lab(self, label):
        if label in self.dic_iri:
            self.countLBS += 1
            iri = self.dic_iri[label]
            lab_org = self.dic_lab[iri]
            type_snips = self.dic_type[iri]
            return iri, lab_org, type_snips
        else:
            self.countNA += 1
            self.untranslated.append(label)
            return 'NA', 'NA', 'NA'
    def print_untrans(self):
        self.untr = list(set(self.untranslated))
        self.untr.sort()
        # print('\t Untranslated terms:')
        # print(f"{Fore.RED}\nUntranslated terms:{Style.RESET_ALL}")
        for item in self.untr:
            print(f"{Fore.RED}\t%s{Style.RESET_ALL}" % item)
            #print('\t', item)
    def reset_untrans(self):
        self.untranslated = []


from owlready2 import *

# def make_phsOnto(onto):
with onto:
    class phenoscript_annotations(AnnotationProperty): pass
    class phs_implies_absence_of(phenoscript_annotations): pass
    class phs_NL(phenoscript_annotations): pass
    class phs_original_assertion(phenoscript_annotations): pass
    class phs_original_class(phenoscript_annotations): pass

phenoscript_annotations.iri     = 'https://github.com/sergeitarasov/PhenoScript/PHS_0000001'
phs_implies_absence_of.iri      = 'https://github.com/sergeitarasov/PhenoScript/PHS_0000003'
phs_NL.iri                      = 'https://github.com/sergeitarasov/PhenoScript/PHS_0000004'
phs_original_assertion.iri      = 'https://github.com/sergeitarasov/PhenoScript/PHS_0000005'
phs_original_class.iri          = 'https://github.com/sergeitarasov/PhenoScript/PHS_0000002'

phenoscript_annotations.label    = 'phenoscript_annotations'
phs_implies_absence_of.label     = 'phs_implies_absence_of'
phs_NL.label                     = 'phs_NL'
phs_original_assertion.label     = 'phs_original_assertion'
phs_original_class.label         = 'phs_original_class'

# Connecting OTUs and traits
with onto:
    class has_trait(AnnotationProperty): pass
    class contributor(AnnotationProperty): pass
    class created_by(AnnotationProperty): pass
    class project_id(AnnotationProperty): pass
    class purl_requires(AnnotationProperty): pass
    class purl_source(AnnotationProperty): pass
    class creation_date(AnnotationProperty): pass
    class cls_project(Thing): pass

has_trait.iri       = 'https://github.com/sergeitarasov/PhenoScript/PHS_0000017'
has_trait.label     = 'has_trait'
contributor.iri     = 'http://purl.org/dc/elements/1.1/contributor'
contributor.label   = 'contributor'
created_by.iri      = 'http://www.geneontology.org/formats/oboInOwl#created_by'
created_by.label    = 'created_by'
project_id.iri      = 'https://github.com/sergeitarasov/PhenoScript/PHS_0000018'
project_id.label    = 'project_id'
purl_requires.iri   = 'http://purl.org/dc/terms/requires'
purl_requires.label = 'requires'
purl_source.iri     = 'http://purl.org/dc/elements/1.1/source'
purl_source.label   = 'source'
cls_project.iri     = 'https://github.com/sergeitarasov/PhenoScript/PHS_0000019'
cls_project.label   = 'project'
creation_date.iri   = 'http://www.geneontology.org/formats/oboInOwl#creation_date'
creation_date.label = 'creation_date'
#  return(onto)


# -----------------------------------------
# Settings for using owlready and Phenoscript
# -----------------------------------------
from phenospy.owl_owlready_config import *
from owlready2 import *
from urllib.parse import urlparse
# -----------------------------------------
# Functions
# -----------------------------------------


# make_Class(onto, iri, label)
def make_Class(onto, iri, label):
     # check existence
    chk = IRIS[iri]
    if chk is None:
        with onto:
            class new_term(Thing): pass
        new_term.iri = iri
        new_term.label = label
        if (new_term.iri != iri):
            print(f"{Fore.RED}Error: cannot set new class iri.{Style.RESET_ALL}")
            exit()

def make_ObjectProperty(onto, iri, label):
     # check existence
    chk = IRIS[iri]
    if chk is None:
        with onto:
            class new_term(ObjectProperty): pass
        new_term.iri = iri
        new_term.label = label
        if (new_term.iri != iri):
            print(f"{Fore.RED}Error: cannot set new class iri.{Style.RESET_ALL}")
            exit()

def make_DataProperty(onto, iri, label):
    # check existence
    chk = IRIS[iri]
    if chk is None:
        with onto:
            class new_dp(DataProperty): pass
        new_dp.iri = iri
        new_dp.label = label
        if (new_dp.iri != iri):
            print(f"{Fore.RED}Error: cannot set new class iri.{Style.RESET_ALL}")
            exit()

def make_AnnotationProperty(onto, iri, label):
    # check existence
    chk = IRIS[iri]
    if chk is None:
        with onto:
            class new_ap(AnnotationProperty): pass
        new_ap.iri = iri
        new_ap.label = label
        if (new_ap.iri != iri):
            print(f"{Fore.RED}Error: cannot set new class iri.{Style.RESET_ALL}")
            exit()


# - get classes
def make_all_classes(root, onto):
    terms_C = root.findall(".//phs:node[@phs:type_onto='C']", ns)
    dic_C = {}
    for i in terms_C:
        #print(i.get(phs + 'iri'), i.get(phs + 'label_original'))
        dic_C.update({i.get(phs + 'iri'): i.get(phs + 'label_original')})
    
    terms_C_props = root.findall(".//phs:node_property[@phs:value-type_onto='C']", ns)
    for i in terms_C_props:
        # print(i.get(phs + 'value-iri'), '>', i.get(phs + 'value-label_original'))
        dic_C.update({i.get(phs + 'value-iri'): i.get(phs + 'value-label_original')})

    for iri in dic_C:
        make_Class(onto, iri=iri, label=dic_C[iri])

    # - get OP
    terms_OP = root.findall(".//phs:edge[@phs:type_onto='OP']", ns)
    dic_OP = {}
    for i in terms_OP:
        dic_OP.update({i.get(phs + 'iri'): i.get(phs + 'label_original')})

    for iri in dic_OP:
        make_ObjectProperty(onto, iri=iri, label=dic_OP[iri])

    # - get DP
    terms_DP = root.findall(".//phs:edge[@phs:type_onto='DP']", ns)
    dic_DP = {}
    for i in terms_DP:
        dic_DP.update({i.get(phs + 'iri'): i.get(phs + 'label_original')})

    for iri in dic_DP:
        # print(iri, dic_DP[iri])
        make_DataProperty(onto, iri=iri, label=dic_DP[iri])
    # - get AP
    terms_AP = root.findall(".//phs:edge[@phs:type_onto='AP']", ns)
    dic_AP = {}
    for i in terms_AP:
        dic_AP.update({i.get(phs + 'iri'): i.get(phs + 'label_original')})

    for iri in dic_AP:
        # print(iri, dic_AP[iri])
        make_AnnotationProperty(onto, iri=iri, label=dic_AP[iri])


# given a list of xml nodes, select unique ones based on attribute attr
def xmlUniqueNodes(nodesXML, attr):
    # attr=phs + 'node_id'
    childID = []
    unique_childrenXML = []
    for child in nodesXML:
        i = child.get(attr)
        if i not in childID:
            childID.append(i)
            unique_childrenXML.append(child)
    return unique_childrenXML

# convert PhenoScript numeric node to porper numbers
def phsNumbersToOWL(str_value, type):
    if(type=='real'):
        out=float(str_value)
    elif(type=='int'):
        out=int(str_value)
    else:
        warnings.warn('Unknown numeric type, see: phsNumbersToOWL()')
    return out

# check is a string is URI
def is_valid_uri(uri_string):
    try:
        parsed_uri = urlparse(uri_string)
        return all([parsed_uri.scheme, parsed_uri.netloc])
    except ValueError:
        return False

#--------------------- Declare new datatype for ontology: xsd:anyURI
class anyURI(object):
    def __init__(self, value):
        self.value = value

def anyURI_parser(s):
    return anyURI(s)

def anyURI_unparser(x):
    return x.value

declare_datatype(anyURI, "http://www.w3.org/2001/XMLSchema#anyURI", anyURI_parser, anyURI_unparser)
#---------------------


from .phs_mainConvert import phsToOWL
from .owl_xmlToOwl import  xmlToOwl
from .phs_grammar  import parsePHS
from .phs_xml import phenoscriptToXML
from .phs_xmlTranslateTerms import xmlTranslateTerms, xmlTranslateTermsMeta
from .snips_makeFromYaml import make_vscodeSnips
from .snips_fun import printUniqueIRIs
from .nl_owlToMd_fun import owlToNLgraph, NLgraphToMarkdown


# cd /Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/phenospy
# /Users/taravser/opt/anaconda3/envs/PhenoScript/bin/python

#
#------------

from colorama import Fore, Style
from colorama import init as colorama_init

colorama_init()

import os
import tempfile
import xml.etree.ElementTree as ET
# from datetime import date
import datetime

from phenospy.owl_xml2owl_fun import *
from phenospy.xml_recodeThis import *
#from xml_recodeThis import *

# -----------------------------------------
# Config: imported from owl_xml2owl_fun
# -----------------------------------------
# from owl_owlready_config import *

# -----------------------------------------
# Arguments
# -----------------------------------------

# phsXML = "/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/phenospy/package-data/phs_xml.xml"
# owl_file = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/phenospy/package-data/test.owl'
# tree = ET.parse(phsXML)
# set_log_level(0)
# xmlToOwl(tree, owl_file)

def xmlToOwl(tree, owl_file):
    print(f"{Fore.BLUE}Converting XML to OWL...{Style.RESET_ALL}")
    # -----------------------------------------
    # Read XML
    # -----------------------------------------
    # tree = ET.parse(phsXML)
    root = tree.getroot()

    # -----------------------------------------
    # Convert 'this' keyword
    # -----------------------------------------
    xmlRecodeThis(root, ns)

    # -----------------------------------------
    # Create ontology
    # -----------------------------------------
    # base_iri = root.find(".//base-iri", ns)
    base_iri = root.find("./phs-config-yaml/base-iri", ns)

    # onto.base_iri = base_iri.text
    base_iri_text = base_iri.text 
    onto = get_ontology(base_iri_text)

    # -----------------------------------------
    # Create classes
    # -----------------------------------------

    make_all_classes(root, onto)

    # -----------------------------------------
    # Save tmp file and then open
    # -----------------------------------------
    # Save the ontology to a temporary file
    # with tempfile.NamedTemporaryFile() as f:
    #     onto.save(file=f.name, format="rdfxml")
    #     # print("Ontology saved to:", f.name)
    #     onto.destroy()
    #     onto = get_ontology(f.name).load()
    # #
    temp = tempfile.NamedTemporaryFile(suffix=".owl", delete=False)
    onto.save(file=temp.name, format="rdfxml")
    temp.close()
    onto.destroy()
    onto = get_ontology(temp.name).load()
    os.unlink(temp.name)
    #
    obo = onto.get_namespace("http://purl.obolibrary.org/obo/")
    phs_ns = onto.get_namespace('https://github.com/sergeitarasov/PhenoScript/')

    # --------------------------------------------------------------------------------
    # Add phenoscript ontology and other terms
    # --------------------------------------------------------------------------------
    module_path = os.path.dirname(__file__)
    owl_file_path = os.path.join(module_path, "owl_make_phsOntology.py")

    with open(owl_file_path, "r") as f_phs:
        phs_make = f_phs.read()
    
    exec(phs_make, {"onto": onto}, globals())

    # -----------------------------------------
    # Add ontology metadata
    # -----------------------------------------
  
    this_Project = cls_project()
    this_Project.iri = base_iri_text + '/project'
    this_Project.project_id.append(base_iri_text)

    yaml_authors = root.find("./phs-config-yaml/authors", ns)
    for author in yaml_authors:
        auct_name = author.get('name') + ' ' + author.text
        onto.metadata.contributor.append(auct_name)
        this_Project.contributor.append(auct_name)

    yaml_ontos = root.find("./phs-config-yaml/importOntologies", ns)
    for i in yaml_ontos:
        onto.metadata.purl_requires.append(i.text)
        this_Project.purl_requires.append(i.text)



    #-----------------------
    # Make nodes
    #-----------------------

    # get unique nodes
    print(f"{Fore.BLUE}Creating nodes...{Style.RESET_ALL}")


    nodeID=[]
    # get unique nodes except those that have phs:fromNegativeEdge='True'
    for node in root.iter(phs+'node'):
        # root.findall(".//phs:node", ns)
        # root.findall(".//phs:node[@phs:fromNegativeEdge='False']", ns)
        is_negEdge = eval(node.get(phs + 'fromNegativeEdge'))
        if is_negEdge==False:
            nodeID.append(node.get(phs + 'node_id'))


    # get unique
    nodeIDunique=list(set(nodeID))



    for ni in nodeIDunique:
        node = root.find(".//phs:node[@phs:node_id='%s']" % ni, ns)

        #print(node)
        nodeClass = node.get(phs + 'iri')
        # this piece of code uses a workaround for iri cuz Owlready does not assign it correctly
        # when using the direct approach newNode = IRIS[nodeClass](ni)
        # create new node
        # newNode = IRIS[nodeClass]('https://temp/tmp')
        nodeClassObj=IRIS[nodeClass]
        newNode = nodeClassObj('https://temp/tmp')
        newNode.iri=ni
        #print(newNode.iri)
        
        className=IRIS[nodeClass].label
        newNode.label = className[0]+':'+ ni.rsplit('/', 1)[-1]

        # add annotations
        newNode.created_by = "phenospy v. " + phsVersion
        # add date
        # today = date.today()
        # dt_string = today.strftime("%d/%m/%Y")
        # newNode.creation_date = dt_string
        newNode.creation_date = datetime.datetime.now()
        #
        # add original class annotation
        newNode.phs_original_class.append(nodeClassObj)
        newNode.purl_source.append(this_Project)
        #
        # add additional/extra class
        extra_classes = node.findall(".//phs:node_property[@phs:value-type_onto='C']", ns)
        if len(extra_classes) > 0:
            # print('Extra CLASS')
            for cl in extra_classes:
                # print(cl.get(phs + 'value-iri'))
                extraClass = IRIS[cl.get(phs + 'value-iri')]
                newNode.is_a.append(extraClass)
                newNode.phs_original_class.append(extraClass)




    #-----------------------
    # Make edges between nodes
    ##  ............................................................................
    ##  Compile N E N: phs:node pos=1 -> E pos=2 -> (phs:node | phs:nested_node | phs:list_node)  pos=3    ####
    #-----------------------

    print(f"{Fore.BLUE}Adding edges between nodes...{Style.RESET_ALL}")

    # make mapping to extract parents quickly
    parent_map = {c:p for p in root.iter() for c in p}
    #parent_map[nodePos1[0]]

    # select node triple_pos=1
    nodePos1=root.findall(".//phs:node[@phs:triple_pos='1']", ns)
    #len(nodePos1)

    #----------------------
    # nodeXML = nodePos1[2]
    for nodeXML in nodePos1:
        parentXML=parent_map[nodeXML]
        edgeXML=parentXML.findall("./*[@phs:triple_pos='2']", ns)
        Pos3XML=parentXML.findall("./*[@phs:triple_pos='3']", ns)
        is_negEdge = eval(edgeXML[0].get(phs + 'negative_prop'))
        #
        if (is_negEdge==False):
            #
            #continue
            #print(is_negEdge)
            #
            if (Pos3XML[0].tag == phs + 'node'):
                # print('ok')
                N1 = IRIS[nodeXML.get(phs + 'node_id')]
                N2 = IRIS[Pos3XML[0].get(phs + 'node_id')]
                Ed = IRIS[edgeXML[0].get(phs + 'iri')]
                if owl.FunctionalProperty in Ed.is_a:
                    exec('N1.%s = N2' % Ed.name)
                    # inf.PhenoScript_original_assertion[N1, Ed, N2] = True
                    phs_original_assertion[N1, Ed, N2] = True
                    # purl_source[N1, Ed, N2] = append(this_Project)
                else:
                    exec('N1.%s.append(N2)' % Ed.name)
                    phs_original_assertion[N1, Ed, N2] = True
                    # purl_source[N1, Ed, N2] = this_Project
                #
            elif (Pos3XML[0].tag == phs + 'list_node'):
                # print('list_node')
                #
                N1 = IRIS[nodeXML.get(phs + 'node_id')]
                Ed = IRIS[edgeXML[0].get(phs + 'iri')]
                childrenXML = Pos3XML[0].findall(".//phs:node[@phs:triple_pos='NO']", ns)

                for child in childrenXML:
                    N2 = IRIS[child.get(phs + 'node_id')]
                    if owl.FunctionalProperty in Ed.is_a:
                        exec('N1.%s = N2' % Ed.name)
                        phs_original_assertion[N1, Ed, N2] = True
                    else:
                        exec('N1.%s.append(N2)' % Ed.name)
                        phs_original_assertion[N1, Ed, N2] = True
                #
            elif (Pos3XML[0].tag == phs + 'nested_node'):
                # print('nested_node')
                #
                N1 = IRIS[nodeXML.get(phs + 'node_id')]
                Ed = IRIS[edgeXML[0].get(phs + 'iri')]
                childrenXML = Pos3XML[0].findall(".//phs:node", ns)
                #
                # get unique nodes based on their ids
                unique_childrenXML = xmlUniqueNodes(childrenXML, attr=phs + 'node_id')
                for child in unique_childrenXML:
                    # print('child: ', child.attrib)
                    N2 = IRIS[child.get(phs + 'node_id')]
                    if owl.FunctionalProperty in Ed.is_a:
                        exec('N1.%s = N2' % Ed.name)
                        phs_original_assertion[N1, Ed, N2] = True
                    else:
                        exec('N1.%s.append(N2)' % Ed.name)
                        phs_original_assertion[N1, Ed, N2] = True
                #
            elif (Pos3XML[0].tag == phs + 'numeric_node'):
                # print(Pos3XML[0].attrib)
                N1 = IRIS[nodeXML.get(phs + 'node_id')]
                Ed = IRIS[edgeXML[0].get(phs + 'iri')]
                N2_type = Pos3XML[0].get(phs + 'numeric_type')
                N2_value = Pos3XML[0].get(phs + 'node_name')
                N2 = phsNumbersToOWL(N2_value, N2_type)
                # [owl.DatatypeProperty, owl.FunctionalProperty] obo.IAO_0000004.is_a
                if owl.FunctionalProperty in Ed.is_a:
                    exec('N1.%s = N2' % Ed.name)
                    phs_original_assertion[N1, Ed, N2] = True
                else:
                    exec('N1.%s.append(N2)' % Ed.name)
                    phs_original_assertion[N1, Ed, N2] = True
            
            elif (Pos3XML[0].tag == phs + 'quoted_node'):
                # print(Pos3XML[0].attrib)
                N1 = IRIS[nodeXML.get(phs + 'node_id')]
                Ed = IRIS[edgeXML[0].get(phs + 'iri')]
                N2 = Pos3XML[0].get(phs + 'node_name')
                if 'http://www.w3.org/2000/01/rdf-schema#label' in Ed.iri:
                    exec('N1.%s = N2' % Ed.name)
                    phs_original_assertion[N1, Ed, N2] = True
                elif (is_valid_uri(N2)):
                    # ins2.new_dp.append(anyURI('https://www.gbif.org/species/10360397'))
                    exec('N1.%s.append(anyURI(N2))' % Ed.name)
                    comment[N1, Ed, anyURI(N2)] = True
                else:
                    exec('N1.%s.append(N2)' % Ed.name)
                    phs_original_assertion[N1, Ed, N2] = True
                #
            else:
                #print('WARNING: something is wrong in "for nodeXML in nodePos1: ... else:"')
                # warnings.warn('something is wrong  with Pos3XML[0].tag in "for nodeXML in nodePos1: ... else:"')
                print(f"{Fore.RED}something is wrong  with Pos3XML[0].tag in for nodeXML in nodePos1: ... else:!{Style.RESET_ALL}")

                #
        elif (is_negEdge == True):
            if (Pos3XML[0].tag == phs + 'node'):
                # print(is_negEdge)
                #
                N1 = IRIS[nodeXML.get(phs + 'node_id')]
                # N2 = IRIS[Pos3XML[0].get(phs + 'node_id')]
                classN2 = IRIS[Pos3XML[0].get(phs + 'iri')]
                Ed = IRIS[edgeXML[0].get(phs + 'iri')]
                #
                # --------------------------------
                # obo.BFO_0000051  # has_part
                # --------------------------------
                # exec('N1.is_a.append( obo.BFO_0000051.some( Not(%s.some(classN2) ) ) )' % Ed)
                # exec('N1.is_a.append(Not(%s.some(classN2)))' % Ed)
                N1.is_a.append(Not(Ed.some(classN2)))
                # add annotation for absence
                # N1.PhenoScript_implies_absence_of.append(classN2)
                N1.phs_implies_absence_of.append(classN2)

                # N2.is_a.append(obo.HAO_0001017) # this works
                # N2.is_a.append(obo.BFO_0000051.some(obo.HAO_0001017)) # this works
                # N2.is_a.append(obo.BFO_0000051.some( Not(obo.BFO_0000051.some(obo.HAO_0001017)) ))  # this works
            else:
                # warnings.warn('Something is wrong with Pos3XML[0].tag in "for nodeXML in nodePos1: ... elif (is_negEdge == True): ...."')
                print(f"{Fore.RED}Something is wrong with 'Pos3XML[0].tag in for nodeXML in nodePos1: ... elif (is_negEdge == True):'!{Style.RESET_ALL}")




    ##  ............................................................................
    ##  Compile N E N: (phs:nested_node | phs:list_node) pos=1 -> E pos=2 -> (phs:node) pos=3    ####

    print(f"{Fore.BLUE}Adding edges between nested and list nodes...{Style.RESET_ALL}")


    # select list or nested nodes, triple_pos=1
    nodePos1=root.findall(".//phs:nested_node[@phs:triple_pos='1']", ns) + root.findall(".//phs:list_node[@phs:triple_pos='1']", ns)

    # nodeXML = nodePos1[0]
    for nodeXML in nodePos1:
        parentXML=parent_map[nodeXML]
        edgeXML=parentXML.findall("./*[@phs:triple_pos='2']", ns)
        Pos3XML=parentXML.findall("./phs:node[@phs:triple_pos='3']", ns)
        is_negEdge = eval(edgeXML[0].get(phs + 'negative_prop'))
        #
        if (Pos3XML[0].tag == phs + 'node' and is_negEdge==False):
            N1_childrenXML = nodeXML.findall(".//phs:node", ns)
            N1_unique_childrenXML = xmlUniqueNodes(N1_childrenXML, attr=phs + 'node_id')
            Ed = IRIS[edgeXML[0].get(phs + 'iri')]
            N2=IRIS[Pos3XML[0].get(phs + 'node_id')]
            # child=N1_unique_childrenXML[0]
            for child in N1_unique_childrenXML:
                N1 = IRIS[child.get(phs + 'node_id')]
                if owl.FunctionalProperty in Ed.is_a:
                    exec('N1.%s = N2' % Ed.name)
                    phs_original_assertion[N1, Ed, N2] = True
                else:
                    exec('N1.%s.append(N2)' % Ed.name)
                    phs_original_assertion[N1, Ed, N2] = True
        else:
            # warnings.warn('something is wrong  with "Pos3XML[0].tag == phs + node and is_negEdge==False" when phs:triple_pos=1 is not a simple node')
            print(f"{Fore.RED}something is wrong  with 'Pos3XML[0].tag == phs + node and is_negEdge==False' when phs:triple_pos=1 is not a simple node!{Style.RESET_ALL}")



    ##  ............................................................................
    ##  Linking OTU and OPHUs    ####

    print(f"{Fore.BLUE}Linking OTU's DATA with TRAITS...{Style.RESET_ALL}")

    OTUs=root.findall(".//phs:otu_object", ns)
    # loop over otu_objects (i.e. species). in each otu_obj identify nodes that link all ophus
    # otu = OTUs[0]
    for otu in OTUs:
        # get all phs:node_property
        props=otu.findall(".//phs:otu_properties//phs:node_property", ns)
        # get all nodes in phs:ophu_list for the otu that are fromNegativeEdge='False'
        ophuList = otu.findall(".//phs:ophu_list", ns)
        ophuNodes_noFil = ophuList[0].findall(".//phs:node[@phs:fromNegativeEdge='False']", ns)
        
        # exclude those that have keyword 'exclude = True'
        ophuNodes = []
        for node in ophuNodes_noFil:
            node_exc = node.find(".//phs:node_property[@phs:var-iri='KeyWord:exclude'][@phs:value-iri='KeyWord:True']", ns)
            if node_exc is None:
                ophuNodes.append(node)
            # else :
            #     print(node_exc.get(phs+'var-iri'))  
        for p in props:
            if (p.text=='linksTraits') and (p.get(phs+'value-iri') == "KeyWord:True"):
                parentXML = parent_map[p]
                N1 = IRIS[parentXML.get(phs+'node_id')]
                # Ed=IRIS[p.get(phs+'iri')]
                Ed = has_trait
                #print(p.text, p.get(phs+'value'), p.get(phs+'iri'), parentXML.get(phs+'node_id'))
                #print(N1, Ed)
                for node in ophuNodes:
                    N2=IRIS[node.get(phs+'node_id')]
                    # if owl.FunctionalProperty in Ed.is_a:
                    #     exec('N1.%s = N2' % Ed)
                    #     phs_original_assertion[N1, Ed, N2] = True
                    # else:
                    exec('N1.%s.append(N2)' % Ed)
                    phs_original_assertion[N1, Ed, N2] = True


    #-----------------------
    # Save file
    #-----------------------
    print(f"{Fore.BLUE}Saving owl file: {Style.RESET_ALL}", owl_file)
    onto.save(file = owl_file, format = "rdfxml")
    print(f"{Fore.GREEN}Done!{Style.RESET_ALL}")



import xml.dom.minidom as xmldom
import xml.etree.ElementTree as ET

from phenospy.phs_addXmlMeta import *
from phenospy.snips_fun import *
from phenospy.snips_readFromJSON import *


def xmlTranslateTerms(root, snipsDic):
  print(f"{Fore.BLUE}Translating terms using JSON snippets...{Style.RESET_ALL}")
  snipsDic.resetCount()
  toExit = False
  # nodes
  for nn in root.iter('{https://github.com/sergeitarasov/PhenoScript}node'):
    # print(nn.get('{https://github.com/sergeitarasov/PhenoScript}node_name'))
    label = nn.get('{https://github.com/sergeitarasov/PhenoScript}node_name')
    iri, lab_org, type_snips = snipsDic.translate_lab(label)
    nn.set('{https://github.com/sergeitarasov/PhenoScript}iri', iri)
    nn.set('{https://github.com/sergeitarasov/PhenoScript}label_original', lab_org)
    nn.set('{https://github.com/sergeitarasov/PhenoScript}type_onto', type_snips)
  
  # print(f"{Fore.GREEN}- Nodes translated: %s{Style.RESET_ALL}" % snipsDic.countLBS)
  if snipsDic.countNA == 0:
    print(f"{Fore.GREEN}- Good! Nodes: all translated!{Style.RESET_ALL}")
  else:
    print(f"{Fore.RED}- Warning! Untranslated nodes found: %s{Style.RESET_ALL}" % snipsDic.countNA)
    toExit = True
    snipsDic.print_untrans()
    snipsDic.reset_untrans()

  # print untranslated terms
  snipsDic.resetCount()

  

  # edges
  for nn in root.iter('{https://github.com/sergeitarasov/PhenoScript}edge'):
    # print(nn.get('{https://github.com/sergeitarasov/PhenoScript}edge_name'))
    label = nn.get('{https://github.com/sergeitarasov/PhenoScript}edge_name')
    iri, lab_org, type_snips = snipsDic.translate_lab(label)
    nn.set('{https://github.com/sergeitarasov/PhenoScript}iri', iri)
    nn.set('{https://github.com/sergeitarasov/PhenoScript}label_original', lab_org)
    nn.set('{https://github.com/sergeitarasov/PhenoScript}type_onto', type_snips)

  # print(f"{Fore.GREEN}- Edges translated: %s{Style.RESET_ALL}" % snipsDic.countLBS)
  # print(f"{Fore.RED}- Edges untranslated: %s{Style.RESET_ALL}" % snipsDic.countNA)
  # snipsDic.resetCount()

  if snipsDic.countNA == 0:
    print(f"{Fore.GREEN}- Good! Edges: all translated!{Style.RESET_ALL}")
  else:
    print(f"{Fore.RED}- Warning! Untranslated edges found: %s{Style.RESET_ALL}" % snipsDic.countNA)
    toExit = True
    snipsDic.print_untrans()
    snipsDic.reset_untrans()
  
  # print untranslated terms
  snipsDic.resetCount()

  
  
  # phs:node_property

  # import itertools
  # nn = next(itertools.islice(root.iter('{https://github.com/sergeitarasov/PhenoScript}node_property'), 2, 3), 
  # None)
  skip_names = ['node_name', 'int', 'numeric_type', 'real']
  # 'int' not in skip_names
  for nn in root.iter('{https://github.com/sergeitarasov/PhenoScript}node_property'):
    # print(nn.text)
    #label=nn.text
    if nn.text not in skip_names:
      # --- text
      # print(nn.text)
      iri, lab_org, type_snips = snipsDic.translate_lab(nn.text)
      nn.set('{https://github.com/sergeitarasov/PhenoScript}var-iri', iri)
      nn.set('{https://github.com/sergeitarasov/PhenoScript}var-label_original', lab_org)
      nn.set('{https://github.com/sergeitarasov/PhenoScript}var-type_onto', type_snips)

      # --- values
      value = nn.get('{https://github.com/sergeitarasov/PhenoScript}value')
      # print(value)
      iri, lab_org, type_snips = snipsDic.translate_lab(value)
      nn.set('{https://github.com/sergeitarasov/PhenoScript}value-iri', iri)
      nn.set('{https://github.com/sergeitarasov/PhenoScript}value-label_original', lab_org)
      nn.set('{https://github.com/sergeitarasov/PhenoScript}value-type_onto', type_snips)
    #----
    # label=nn.get('{https://github.com/sergeitarasov/PhenoScript}value')
    # iri=iriDic.translate(label)
    # #print(iri)
    # nn.set('{https://github.com/sergeitarasov/PhenoScript}iri', iri)
  
  # print('- translated phs:node_property: %s; NA phs:node_property: %s.' % (iriDic.countLBS, iriDic.countNA))
  # iriDic.resetCount()
  if snipsDic.countNA == 0:
    print(f"{Fore.GREEN}- Good! Node Properties: all translated!{Style.RESET_ALL}")
  else:
    print(f"{Fore.RED}- Warning! Untranslated Node Properties found: %s{Style.RESET_ALL}" % snipsDic.countNA)
    toExit = True
    snipsDic.print_untrans()
    snipsDic.reset_untrans()

  #
  snipsDic.resetCount()

  if toExit == True:
    print(f"{Fore.RED}Warning! Fix untralstated items and re-run.{Style.RESET_ALL}")
    exit()






# -----------------------------------------
# Translate XML and add metadata
# -----------------------------------------
def xmlTranslateTermsMeta(xml_str, base_iri, yaml_file):
    # -----------------------------------------
    # Read yaml
    # -----------------------------------------
    print(f"{Fore.BLUE}Reading yaml file...{Style.RESET_ALL}")
    with open(yaml_file, 'r') as f_yaml:
        phs_yaml = yaml.safe_load(f_yaml)
    # print(phs_yaml)
    print(f"{Fore.GREEN}Good! File is read!{Style.RESET_ALL}")

    # -----------------------------------------
    # Prepearing snippets
    # -----------------------------------------
    print(f"{Fore.BLUE}Preparing snippets for translation...{Style.RESET_ALL}")
    dic_iri, dic_type, dic_lab = make_dicsFromSnips(phs_yaml)
    snipsDic = snipsLabelDictionary(dic_iri, dic_type, dic_lab)

    # -----------------------------------------
    # Prepearing XML and translating
    # -----------------------------------------
    tree = ET.ElementTree(ET.fromstring(xml_str))
    ET.register_namespace('phs', 'https://github.com/sergeitarasov/PhenoScript')
    root = tree.getroot()

    print(f"{Fore.BLUE}Translating XML...{Style.RESET_ALL}")
    xmlTranslateTerms(root, snipsDic)

    # -----------------------------------------
    # Add Meta
    # -----------------------------------------
    print(f"{Fore.BLUE}Adding metadata to XML...{Style.RESET_ALL}")
    #tree = ET.ElementTree(ET.fromstring(root))
    #ET.register_namespace('phs', 'https://github.com/sergeitarasov/PhenoScript')
    # root = tree.getroot()
    xml_add_YamlMeta(root, phs_yaml, base_iri)
    
    # -----------------------------------------
    # Pretty-print of the combined XML tree
    # -----------------------------------------
    xml_string = ET.tostring(root, encoding='unicode')
    xml_string = xmldom.parseString(xml_string).toprettyxml(indent="\t", newl="\n", encoding=None)
    xml_string = "\n".join([ll.rstrip() for ll in xml_string.splitlines() if ll.strip()])
    print(f"{Fore.GREEN}Done!{Style.RESET_ALL}")
    return xml_string


# -----------------------------------------------------------
# various handy functions to translate phs into xml
#
# (C) 2023 Sergei Tarasov
# email sergei.tarasov@helsinki.fi
# -----------------------------------------------------------
import uuid

from phenospy.snips_fun import *


# print(uuid_n(6))  # For example, da86ce
def uuid_n(string_length=6):
    """Returns a random string of length string_length."""
    random = str(uuid.uuid4())
    # random = random.upper()
    random = random.replace("-", "")
    return random[0:string_length]


# --------- auto ids for nodes
# nid = nodeIdGenerator(prefix='xx', starting_id=1)
# nid
# nid.makeId()
class nodeIdGenerator:
    def __init__(self, prefix='', starting_id=1):
        self.current_id = starting_id
        self.out_id = 0
        self.prefix = prefix
    def makeId(self):
        self.out_id = self.current_id
        self.out_id = self.prefix + str(self.out_id)
        self.current_id += 1
        return self.out_id
    def reset(self, prefix='', starting_id=1):
        self.current_id = starting_id
        self.out_id = 0
        self.prefix = prefix
    def __repr__(self):
        return str(self.out_id)


# -----------------------------------------
# This functions check is parentheses are balanced in a string
# from: https://stackoverflow.com/questions/38833819/python-program-to-check-matching-of-simple-parentheses/38834005
# is_balanced(str)
# -----------------------------------------
def is_balanced(expr):
    # print('Checking if parentheses are balanced...', flush=True)
    print(f"{Fore.BLUE}Checking if parentheses are balanced...{Style.RESET_ALL}")
    opening = set('([{')
    new = set(')]}{[(')
    match = set([('(', ')'), ('[', ']'), ('{', '}')])
    stack = []
    stackcount = []
    for i, char in enumerate(expr, 1):
        if char not in new:
            continue
        elif char in opening:
            stack.append(char)
            stackcount.append(i)
        else:
            if len(stack) == 0:
                # print(i)
                # print('Some parentheses might be disbalanced, see line:', txt[1:i].count('\n') + 1, flush=True)
                print(f"{Fore.RED}Some parentheses might be disbalanced, see line:{Style.RESET_ALL}", expr[1:i].count('\n') + 1)
                return False
            lastOpen = stack.pop()
            lastindex = stackcount.pop()
            if (lastOpen, char) not in match:
                # print (i)
                # print('Some parentheses might be disbalanced, see line:', expr[1:i].count('\n') + 1, flush=True)
                print(f"{Fore.RED}Some parentheses might be disbalanced, see line:{Style.RESET_ALL}", expr[1:i].count('\n') + 1)
                return False
    length = len(stack)
    if length != 0:
        elem = stackcount[0]
        # print(elem, flush=True)
        print(f"{Fore.RED}Some parentheses might be disbalanced, see:{Style.RESET_ALL}", elem)
    # return length == 0
    if (length == 0):
        print(f"{Fore.GREEN}Good! Parentheses are balanced.{Style.RESET_ALL}")
    else:
        print(f"{Fore.RED}Some parentheses might be disbalanced!{Style.RESET_ALL}")


from colorama import Fore, Style
from colorama import init as colorama_init
colorama_init()
from phenospy.owl_owlready_config import *
from phenospy.nl_owlToMd_fun import *

# -----------------------------------------
# Arguments
# -----------------------------------------
owl_file = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/examples/Grebennikovius/output/Grebennikovius.owl'

# -----------------------------------------
# Making NL graph
# -----------------------------------------
onto = owlToNLgraph(owl_file)

# -----------------------------------------
# NL graph to Markdown
# -----------------------------------------
fl_md = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/devel/test1.md'
ind0 = onto.search(label = 'org_Grebennikovius_basilewskyi')[0]
NLgraphToMarkdown(onto, ind0, file_save = fl_md, verbose =True)

fl_md = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/devel/test2.md'
ind0 = onto.search(label = 'org_Grebennikovius')[0]
NLgraphToMarkdown(onto, ind0, file_save = fl_md, verbose =True)


import xml.etree.ElementTree as ET


# tree = ET.ElementTree(ET.fromstring(out_xml))
# ET.register_namespace('phs', 'https://github.com/sergeitarasov/PhenoScript')
# root = tree.getroot()
# xml_add_YamlMeta(root, phs_yaml, base_iri)
def xml_add_YamlMeta(root, phs_yaml, base_iri):

    data = phs_yaml
    # create the root XML element
    meta = ET.Element("phs-config-yaml")

    # add the base-iri element
    iri_elem = ET.SubElement(meta, "base-iri")
    iri_elem.text = base_iri

    # add the authors element and its subelements
    authors = ET.SubElement(meta, "authors")
    for name, orcid in data["authors"].items():
        author = ET.SubElement(authors, "author")
        author.set("name", name)
        author.text = orcid

    # add the importOntologies element and its subelements
    import_ontologies = ET.SubElement(meta, "importOntologies")
    for ontology in data["importOntologies"]:
        import_ontology = ET.SubElement(import_ontologies, "importOntology")
        import_ontology.text = ontology

    return root.insert(0, meta)








# former namespace_functions.py

import json
import os
from sys import exit

import yaml
from colorama import Fore, Style
from colorama import init as colorama_init
from owlready2 import *

# import importlib.resources as importlib_resources

colorama_init()
#print(f"This is {Fore.GREEN}color{Style.RESET_ALL}!")

#----


# returns Path to phenospy package
def get_phenospyPath():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return dir_path

# iri = 'http://purl.obolibrary.org/obo/HAO_0000232'
# pref, trimmed_iri = getNamespace(iri, nmp_store)
def getNamespace(iri, nmp_store):
    for i in nmp_store:
        #print(i[2])
        spl = iri.split(i[2])
        #print(spl)
        if (len(spl) == 2) and (spl[0] == ''):
            return (i[1], spl[1])
    return (None, None)

# pref, trimmed_iri = getNamespace('http://purl.obolibrary.org/obo/IAO_8000011', nmp_store)
# snip = snippetInfo('C', 'this label', 'http://purl.obolibrary.org/obo/IAO_8000011', pref,  trimmed_iri)
# snip = snippetInfo('C', 'dd', 'http://purl.obolibrary.org/obo/IAO_8000011', 'Default',  trimmed_iri)
# print(snip)
# print(snip.printJson())
class snippetInfo:
    def __init__(self, type, label = None, iri = None, prefix = None, trimmed_iri = None, definition='', sep='-'):
        self.type = type
        self.label = label
        self.iri = iri
        self.prefix = prefix
        self.trimmed_iri = trimmed_iri
        self.label_phs = None
        # fixing defintion
        self.definition = definition
        self.definition =  self.definition.replace('\n', '  ')
        self.definition = self.definition.replace('\t', '  ')
        self.definition = self.definition.replace('"', "'")
        self.definition = '(' + self.type + '): ' + self.definition
        if (iri is None):
            print('iri  is absent')
        if  (label is None):
            self.label = trimmed_iri
        if (prefix is None):
            print('prefix is absent')
            self.prefix = 'NOPREF'
        self.label_phs = self.prefix + sep + self.label
        self.iri4snippet = self.prefix + sep + trimmed_iri
        if  (prefix == 'Default'):
            self.prefix = 'Default'
            self.label_phs = self.label
            self.iri4snippet = 'def' + sep +trimmed_iri
        self.label_phs = self.label_phs.replace(' ', '_')
        self.label_phs = self.label_phs.replace('/', '_')
        self.label_4_translation = self.label_phs
        self.label_no_ns = self.label.replace(' ', '_')
        #self.label_no_ns = self.label_phs
        if (type == 'OP') or (type == 'DP') or (type == 'AP'):
            self.label_phs = '.' + self.label_phs
    def printJson(self):
        out = \
        '"(%s) %s": {\n' \
        '"prefix": ["%s"],\n' \
        '"body": ["%s"],\n' \
        '"description": "%s",\n' \
        '"iri": "%s",\n' \
        '"label_4_translation": ["%s"],\n' \
        '"label_original": "%s",\n' \
        '"type": "%s"\n' \
        '}' % (self.type, self.iri4snippet, self.label_phs, self.label_phs, self.definition, self.iri, \
            self.label_4_translation, self.label, self.type)
        return str(out)
    def __repr__(self):
        ret = 'type: ' + str(self.type) + '\n' +\
            '\t original label: ' + str(self.label) + '\n' + \
            '\t label phs-no-ns: ' + str(self.label_no_ns) + '\n' + \
            '\t label phs: ' + str(self.label_phs)  + '\n' + \
            '\t label 4 translation: ' + str(self.label_4_translation) + '\n' + \
            '\t iri: ' + str(self.iri) + '\n' + \
            '\t namespace pref: ' + str(self.prefix) + '\n' +\
            '\t trimmed_iri: ' + str(self.trimmed_iri)  + '\n' +\
            '\t definition: ' + str(self.definition) + '\n' + \
            '\t iri 4 snippet: ' + str(self.iri4snippet) + '\n'
        return str(ret)

# yaml_file = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/src/reasoning/phs-config.yaml'
# printUniqueIRIs(yaml_file)
def printUniqueIRIs(yaml_file):
    # -----------------------------------------
    # Read configuration yaml
    # -----------------------------------------
    print(f"{Fore.BLUE}Reading yaml file...{Style.RESET_ALL}")
    with open(yaml_file, 'r') as f_yaml:
        phs_yaml = yaml.safe_load(f_yaml)
    # print(phs_yaml)
    print(f"{Fore.GREEN}Good! File is read!{Style.RESET_ALL}")
    #
    # -----------------------------------------
    # Processing ontologies
    # -----------------------------------------
    print(f"{Fore.BLUE}Start reading ontologies...{Style.RESET_ALL}")
    # get ontologies from yaml
    onto_store = phs_yaml['importOntologies']
    # keep all snips here:
    iris_underscore = []
    iris_hash = []
    for onto_path in onto_store:
        # print(onto_path)
        print(f"\n\t{Fore.BLUE}Loading ontology:{Style.RESET_ALL}", onto_path)
        try:
            onto = get_ontology(onto_path).load()
        except:
            onto = get_ontology(onto_path).load()
        obo = onto.get_namespace("http://purl.obolibrary.org/obo/")
        print(f"\t\t{Fore.GREEN}The ontology loaded!{Style.RESET_ALL}")
        # --------------
        # Classes (C)
        # --------------
        # x = 'http://purl.obolibrary.org/obo/pato#increased_in_magnitude_relative_to'
        # x.split('_')

        print(f"\t{Fore.BLUE}Extracting IRIs...{Style.RESET_ALL}")
        #
        onto_obj = onto.classes()
        for item in onto_obj:
            i = item.iri.split('_')[0]
            iris_underscore.append(i)
            if '#' in item.iri:
                i = item.iri.split('#')[0]
                iris_hash.append(i)

        # --------------
        # OP
        # --------------
        onto_obj = onto.object_properties()
        for item in onto_obj:
            i = item.iri.split('_')[0]
            iris_underscore.append(i)
            if '#' in item.iri:
                i = item.iri.split('#')[0]
                iris_hash.append(i)
        # --------------
        # DP
        # --------------
        onto_obj = onto.data_properties()
        for item in onto_obj:
            i = item.iri.split('_')[0]
            iris_underscore.append(i)
            if '#' in item.iri:
                i = item.iri.split('#')[0]
                iris_hash.append(i)
        # --------------
        # AP
        # --------------
        onto_obj = onto.annotation_properties()
        for item in onto_obj:
            i = item.iri.split('_')[0]
            iris_underscore.append(i)
            if '#' in item.iri:
                i = item.iri.split('#')[0]
                iris_hash.append(i)
        #
        print(f"\t{Fore.GREEN}IRIs extracted!\n{Style.RESET_ALL}")
    #
    unique_underscore = set(iris_underscore)
    unique_hash = set(iris_hash)
    print(f"\t{Fore.RED}\nUnique IRIs split with '_':{Style.RESET_ALL}")
    print(*sorted(unique_underscore), sep='\n')
    print(f"\t{Fore.RED}\nUnique IRIs split with '#':{Style.RESET_ALL}")
    print(*sorted(unique_hash), sep ='\n')
#





from setuptools import setup, find_packages

setup(
    name='phenospy',
    version='0.197',
    packages=find_packages(exclude=['tests*', 'devel*', 'build*']),
    license='MIT',
    description='Tools for making and processing computable phenotype descriptions',
    long_description_content_type="text/markdown",
    long_description=open('README.md').read(),
    install_requires=['Owlready2',
                      'PyYAML',
                      'colorama',
                      'pyparsing==2.4.2'],
    include_package_data=True,
    url='https://github.com/sergeitarasov/PhenoScript',
    author='Sergei Tarasov',
    author_email='sergei.tarasov@helsinki.fi',
    project_urls={
        "Project Logo": "https://raw.githubusercontent.com/sergeitarasov/PhenoScript/main/Phenoscript_logo.png",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    
)


import xml.etree.ElementTree as ET

# -----------------------------------------
# Recoding 'this' keyword to respective nodes 
# -----------------------------------------

def xmlRecodeThis(root, ns):
    phs = '{https://github.com/sergeitarasov/PhenoScript}'
    # make mapping to extract parents quickly
    parent_map = {c:p for p in root.iter() for c in p}
    OTUs=root.findall(".//phs:otu_object", ns)
    for otu in OTUs:
        this_node=otu.find(".//phs:otu_properties//phs:node_property[@phs:var-iri='KeyWord:this'][@phs:value-iri='KeyWord:True']", ns)
        # print('THIS:', this_node)
        if this_node is not None:
            this_parent = parent_map[this_node]
            # <phs:node phs:node_id="https://urn:uuid:8f7fbcbc-ccbb-11ed-beb4-acde48001122/id-4e3742" phs:triple_pos="3" phs:node_name="hao-female_organism" 
            # phs:fromNegativeEdge="False" phs:iri="http://purl.obolibrary.org/obo/HAO_0000028" phs:label_original="female organism" phs:type_onto="C">
            p_node_id         = this_parent.get(phs+'node_id')
            p_node_name       = this_parent.get(phs+'node_name')
            p_iri             = this_parent.get(phs+'iri')
            p_label_original  = this_parent.get(phs+'label_original')
            p_type_onto       = this_parent.get(phs+'type_onto')
            #
            ophuList = otu.find(".//phs:ophu_list", ns)
            this_ophuNodes = ophuList.findall(".//phs:node[@phs:iri='KeyWord:this']", ns)
            for this in this_ophuNodes:
                this.set('{https://github.com/sergeitarasov/PhenoScript}node_id', p_node_id)
                this.set('{https://github.com/sergeitarasov/PhenoScript}node_name', p_node_name)
                this.set('{https://github.com/sergeitarasov/PhenoScript}iri', p_iri)
                this.set('{https://github.com/sergeitarasov/PhenoScript}label_original', p_label_original)
                this.set('{https://github.com/sergeitarasov/PhenoScript}type_onto', p_type_onto)
                # this.text 

# Save
# Create an ElementTree object from the root element
# tree_save = ET.ElementTree(root)
# # Save the XML tree to a file
# xml_save = '/Users/taravser/Documents/My_papers/PhenoScript_main/PhenoScript/phenospy_package/phenospy/package-data' \
#            '/phs_xml.xml'
# tree_save.write(xml_save, encoding="utf-8", xml_declaration=True)